# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
import threading
import time
from random import randint

import zmq

from .common import CommonTransport

log = logging.getLogger(__name__)


def _encoded(func):
    def encoded(self, input, *args):
        decoded = input.decode('utf-8')
        deserialized = json.loads(decoded)

        try:
            retval = func(self, deserialized, *args)
            response_string = json.dumps(retval)
        except ValueError as e:
            response_string = str(e)
            log.error(str(e))
        out = response_string.encode('utf-8')
        return out
    return encoded


HEARTBEAT_LIVENESS = 3
HEARTBEAT_INTERVAL = 1
INTERVAL_INIT = 1
INTERVAL_MAX = 32

PPP_READY = b'\x01'
PPP_HEARTBEAT = b'\x02'
PPP_SHUTDOWN = b'\x03'
PPP_REQUEST = b'\x04'
PPP_RESPONSE = b'\x05'

PREFIX = randint(0, 0x10000)


LISTEN_MESSAGE_RESULT_TIMEOUT = 0
LISTEN_MESSAGE_RESULT_INVALID = 1
LISTEN_MESSAGE_RESULT_HEARTBEAT = 2
LISTEN_MESSAGE_RESULT_HANDLED_REQUEST = 3


def worker_socket(addr, context, poller):
    socket = context.socket(zmq.DEALER)
    identity = "%04X-%04X" % (PREFIX, randint(0, 0x10000))
    socket.setsockopt_string(zmq.IDENTITY, identity)
    poller.register(socket, zmq.POLLIN)
    socket.connect(addr)
    socket.send(PPP_READY)
    return socket


class Worker(threading.Thread, CommonTransport):
    """
    Worker is a Paranoid-Pirate worker described in the zguide:
    Related RFC: https://rfc.zeromq.org/spec:6/PPP
    refs:
    http://zguide.zeromq.org/py:all#Robust-Reliable-Queuing-Paranoid-Pirate-Pattern
    """
    def __init__(self, z_context, addr, stopper, registry=None):
        threading.Thread.__init__(self)
        CommonTransport.__init__(self, registry)
        self.addr = addr
        self.z_context = z_context
        self.stopper = stopper
        self.bounce_count = -1
        self.request_id = None
        self.busy_lock = threading.Lock()

    def generate_request_id(self):
        prefix = self.socket_name[5:]
        request_id = "%s-%04X" % (prefix, randint(0, 0x10000))
        return request_id.encode('utf8')

    def setup_zmq_sockets(self):
        self.poller = zmq.Poller()
        self.socket = worker_socket(self.addr, self.z_context, self.poller)
        self.socket_name = self.socket.getsockopt_string(zmq.IDENTITY)
        self.liveness = HEARTBEAT_LIVENESS
        self.interval = INTERVAL_INIT
        self.heartbeat_at = time.time() + HEARTBEAT_INTERVAL

    def run(self):
        self.setup_zmq_sockets()
        self.run_message_loop()

    # This method handle messages from sockets:
    # 1. It performs heartbeat
    # 2. It takes request and handle them
    # 3. It returns value when there is a response from server to plugin
    #
    # (3) should only occurs when this method is called recursively
    # Return values:
    # None - The socket is closing
    # Otherwise - response from server
    #
    # Btw, I cannot think of a better name for this method,
    # please update this if you have a better name
    def run_message_loop(self, send_heartbeat=True):
        timeout = (HEARTBEAT_INTERVAL * 1000) if send_heartbeat else None
        while True:
            result = self.listen_to_message_once(timeout)
            if result == LISTEN_MESSAGE_RESULT_INVALID:
                return None
            elif result == LISTEN_MESSAGE_RESULT_HEARTBEAT:
                self.liveness = HEARTBEAT_LIVENESS
            elif result == LISTEN_MESSAGE_RESULT_TIMEOUT:
                self.handle_heartbeat_timeout()

            if send_heartbeat and time.time() > self.heartbeat_at:
                self.send_heartbeat()

            if result not in [LISTEN_MESSAGE_RESULT_HEARTBEAT,
                              LISTEN_MESSAGE_RESULT_TIMEOUT,
                              LISTEN_MESSAGE_RESULT_HANDLED_REQUEST]:
                # We have a response
                return result

            if self.stopper.is_set():
                self.shutdown_socket()
                return None

    def listen_to_message_once(self, timeout=None):
        poller = self.poller
        socks = dict(poller.poll(timeout))
        if socks.get(self.socket) != zmq.POLLIN:
            return LISTEN_MESSAGE_RESULT_TIMEOUT

        #  Get message
        #  - 7-part envelope + content -> request
        #  - 1-part HEARTBEAT -> heartbeat
        frames = self.socket.recv_multipart()
        if len(frames) == 7:
            client = frames[0]
            assert frames[1] == b''
            message_type = frames[2]
            bounce_count = int(frames[3].decode('utf8'))
            request_id = frames[4]
            assert frames[5] == b''
            message = frames[6]

            self.request_id = request_id
            self.bounce_count = bounce_count
            self.liveness = HEARTBEAT_LIVENESS

            if message_type == PPP_REQUEST:
                self.mark_as_busy_if_needed()
                ctx = {
                    'bounce_count': bounce_count,
                    'request_id': request_id.decode('utf8'),
                }
                response = self.handle_message(message, ctx)
                self.socket.send_multipart([
                    client,
                    b'',
                    PPP_RESPONSE,
                    str(bounce_count).encode('utf8'),
                    request_id,
                    b'',
                    response,
                ])
                self.mark_as_not_busy_if_needed()
                return LISTEN_MESSAGE_RESULT_HANDLED_REQUEST
            elif message_type == PPP_RESPONSE:
                self.mark_as_not_busy_if_needed()
                self.bounce_count -= 1
                return message.decode('utf8')
        elif len(frames) == 1 and frames[0] == PPP_HEARTBEAT:
            return LISTEN_MESSAGE_RESULT_HEARTBEAT
        else:
            log.warn(
                'Invalid message: %s, assuming socket dead', frames)
            return LISTEN_MESSAGE_RESULT_INVALID

    def mark_as_busy_if_needed(self):
        if self.bounce_count == 0:
            self.busy_lock.acquire()

    def mark_as_not_busy_if_needed(self):
        if self.bounce_count == 0:
            self.busy_lock.release()

    def handle_heartbeat_timeout(self):
        self.liveness -= 1
        if self.liveness == 0:
            log.warn('Heartbeat failure, can\'t reach queue')
            log.warn('Reconnecting in %0.2fs...' % self.interval)
            time.sleep(self.interval)

            if self.interval < INTERVAL_MAX:
                self.interval *= 2
            self.shutdown_socket(send_shutdown=False)
            self.socket = worker_socket(
                self.addr,
                self.z_context,
                self.poller
            )
            self.socket_name = self.socket.getsockopt_string(
                zmq.IDENTITY
            )
            self.liveness = HEARTBEAT_LIVENESS
            self.interval = INTERVAL_INIT

    def send_heartbeat(self):
        self.heartbeat_at = time.time() + HEARTBEAT_INTERVAL
        self.socket.send(PPP_HEARTBEAT)

    def shutdown_socket(self, send_shutdown=True):
        if send_shutdown:
            self.socket.send(PPP_SHUTDOWN)
        self.poller.unregister(self.socket)
        self.socket.setsockopt(zmq.LINGER, 0)
        self.socket.close()

    @_encoded
    def handle_message(self, req, extraContext={}):
        kind = req.get('kind')
        if kind == 'init':
            raise Exception('Init trigger is deprecated, '
                            'use init event instead')

        name = req.get('name')
        param = req.get('param', {})
        ctx = req.get('context', {})
        ctx.update(extraContext)

        if kind == 'provider':
            action = param.pop('action')
            return self.call_provider(ctx, name, action, param)
        elif kind == 'handler':
            return self.call_handler(ctx, name, param)
        elif kind == 'event':
            return self.call_event_func(name, param)
        else:
            return self.call_func(ctx, kind, name, param)

    def send_action(self, action_name, payload):
        if self.request_id is None:
            # Sending non-nested request
            self.request_id = self.generate_request_id()
        self.bounce_count += 1
        self.mark_as_busy_if_needed()
        message = {
            'method': 'POST',
            'payload': payload,
        }
        self.socket.send_multipart([
            self.socket_name.encode('utf8'),
            b'',
            PPP_REQUEST,
            str(self.bounce_count).encode('utf8'),
            self.request_id,
            b'',
            json.dumps(message).encode('utf8')
        ])
        return self.run_message_loop(send_heartbeat=False)


class OneOffWorker(Worker):

    def __init__(self, z_context, addr, stopper, registry=None,
                 action_name=None, payload={}):
        Worker.__init__(self, z_context, addr, stopper, registry=None)
        self.action_name = action_name
        self.payload = payload

    def run(self):
        self.setup_zmq_sockets()
        self._result = self.send_action(self.action_name, self.payload)
        self.shutdown_socket()

    def join(self):
        threading.Thread.join(self)
        return self._result


class ZmqTransport(CommonTransport):
    """
    ZmqTransport will start the working thread which run it own zmq socket.
    Since the zmq socket is not thread safe, the worker will be responabile for
    doing their own heartbeat to the keep alive. To skygear-server it just
    like multiple worker process.
    """

    def __init__(self, addr, context=None,
                 registry=None, threading=4, limit=10):
        super().__init__(registry)
        if context is None:
            context = zmq.Context()

        self._addr = addr
        self._context = context
        self._threading = threading
        self.threads = []
        self.threads_opened = 0
        self.limit = limit

    def run(self):
        self.start()
        try:
            self.maintain_workers_count()
        except KeyboardInterrupt:
            log.info('Shutting down all worker')
            self.stop()

    def start(self):
        self.stopper = threading.Event()
        for i in range(self._threading):
            self.start_worker_at_index(i)

    def maintain_workers_count(self):
        i = 0
        while True:
            t = self.threads[i]
            t.join(HEARTBEAT_INTERVAL * HEARTBEAT_LIVENESS)
            if self.stopper.is_set():
                log.info(
                    'Workers are shutting down, stop maintain workers loop')
                return
            if not t.is_alive():
                log.warn(
                    'Worker Thread dead, starting a new one')
                self.start_worker_at_index(i)
            i = i + 1
            if i >= self._threading:
                i = 0
            busy_count = len(
                [t for t in self.threads if t.busy_lock.locked()]
            )
            if busy_count == self._threading and busy_count < self.limit:
                self.start_worker_at_index(self._threading)
                self._threading += 1

    def start_worker_at_index(self, index):
        t = Worker(self._context, self._addr, self.stopper)
        if index < len(self.threads):
            self.threads[index] = t
        else:
            self.threads.append(t)
        t.start()
        self.threads_opened = self.threads_opened + 1

    def stop(self):
        self.stopper.set()
        for t in self.threads:
            t.join()

    def send_action(self, action_name, payload, url=None, timeout=60):
        worker = threading.current_thread()
        if isinstance(worker, Worker):
            # Nested request
            result = worker.send_action(action_name, payload)
            return json.loads(result)

        new_worker = OneOffWorker(
            self._context,
            self._addr,
            self.stopper,
            action_name=action_name,
            payload=payload,
        )
        new_worker.start()
        result = new_worker.join()
        return json.loads(result)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..error import SkygearException, InvalidArgument
from ..options import options as skygear_options
from .fs import FileSystemAssetSigner
from .s3 import S3AssetSigner
from .cloud import CloudAssetSigner

signer = None


def get_signer():
    global signer

    if signer and signer.available():
        return signer

    store_type = skygear_options.asset_store
    if store_type == 'fs':
        signer = FileSystemAssetSigner.create(skygear_options)
    elif store_type == 's3':
        signer = S3AssetSigner.create(skygear_options)
    elif store_type == 'cloud':
        signer = CloudAssetSigner.create(skygear_options)
    else:
        raise SkygearException(
            'Unknown asset store type: {}'.format(store_type),
            code=InvalidArgument
        )

    return signer



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

import bcrypt
import pytest
from sqlalchemy.sql import text

from skygear.container import SkygearContainer
from skygear.utils import user as u
from skygear.utils import db

PLAINTEXT = 'helloworld!'


def assert_correct_pw(password, salt):
    decoded_salt = salt.encode('utf-8')
    new_hash = bcrypt.hashpw(password.encode('utf-8'), decoded_salt)
    assert new_hash == decoded_salt


class TestHashPassword():
    def test_hash_password(self):
        hashed = u.hash_password(PLAINTEXT)
        assert isinstance(hashed, str)
        assert_correct_pw(PLAINTEXT, hashed)


class TestResetPassword(unittest.TestCase):
    app_name = '_'

    def setUp(self):
        SkygearContainer.set_default_app_name(self.app_name)
        with db.conn() as conn:
            conn.execute("CREATE SCHEMA IF NOT EXISTS app_{0}"
                         .format(self.app_name))
            conn.execute("""
                CREATE TABLE IF NOT EXISTS _user (
                    id text PRIMARY KEY,
                    username text,
                    email text,
                    password text,
                    auth jsonb
                );""")
            sql = text("""
                INSERT INTO _user (id, username, password)
                VALUES (:id, :username, :password);
                """)
            conn.execute(sql,
                         id='1',
                         username='USER_1',
                         password=u.hash_password('supersecret1'))

    def tearDown(self):
        with db.conn() as conn:
            conn.execute("DROP TABLE app_{0}._user;".format(self.app_name))

    def test_reset_password(self):
        done = u.reset_password_by_username("USER_1", PLAINTEXT)
        assert done

        with db.conn() as conn:
            result = conn.execute(text("""
                SELECT password
                FROM app_{0}._user
                WHERE username=:username
                """.format(self.app_name)),
                username='USER_1')
            r = result.fetchone()
            assert_correct_pw(PLAINTEXT, r[0])

    def test_no_such_user(self):
        done = u.reset_password_by_username("USER_2", PLAINTEXT)
        assert not done

    def test_bad_parameter(self):
        with pytest.raises(ValueError):
            u.reset_password_by_username(1, '')

        with pytest.raises(ValueError):
            u.reset_password_by_username('', 1)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def push_device(container, device_id, notification, topic=None):
    return push_devices(container, [device_id], notification, topic)


def push_devices(container, device_ids, notification, topic=None):
    params = {
            'device_ids': device_ids,
            'notification': notification,
            }

    if topic:
        params['topic'] = topic

    return container.send_action('push:device', params)


def push_user(container, user_id, notification, topic=None):
    return push_users(container, [user_id], notification, topic)


def push_users(container, user_ids, notification, topic=None):
    params = {
            'user_ids': user_ids,
            'notification': notification,
            }

    if topic:
        params['topic'] = topic

    return container.send_action('push:user', params)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import importlib
import os
from importlib.machinery import SourceFileLoader

from .settings import settings as module_settings
from .settings.module import _config_module, add_module


class LoadException(Exception):
    pass


def guess_package_name(path):
    """
    Guess the package name for the specified file path.
    """
    relpath = os.path.relpath(os.path.expanduser(path))
    if relpath.startswith('..'):
        msg = "Path {0} must reside in current working directory".format(path)
        raise LoadException(msg)

    if relpath == '.' or relpath == '__init__.py':
        return 'plugin'

    if os.path.basename(relpath) == '__init__.py':
        # abc/def/__init__.py -> abc.def
        return os.path.dirname(relpath).replace(os.sep, '.')

    if os.path.isfile(relpath):
        # abc/def/plugin.py -> abc.def.plugin
        return os.path.dirname(relpath).replace(os.sep, '.') + '.' + \
            os.path.splitext(os.path.basename(relpath))[0]

    if os.path.isdir(relpath):
        # abc/def/plugin -> abc.def.plugin
        return relpath.replace(os.sep, '.')

    msg = "Unable to load {0}.".format(path)
    raise LoadException(msg)


def load_module(path):
    """
    Load the module specified by path.

    Path can be one of the followings:
    - If path is a file, load the file as a python module.
    - If path is a directory, load the file as a python package. The directory
      must contains `__init__.py`
    - Else, treat path as the name of the package to load.

    For file, we assume it is cloud code and have no `includeme` function.

    For directory or package, we regards it is explicit imported module, so we
    will config the module once loaded. i.e. calling the `includeme` function
    of the module.
    """
    if os.path.isfile(path):
        package_name = guess_package_name(path)
        loaded = SourceFileLoader(package_name, path).load_module()
        add_module(package_name, loaded)
    elif os.path.isdir(path):
        package_name = guess_package_name(path)
        loadpath = os.path.join(path, '__init__.py')
        loaded = SourceFileLoader(package_name, loadpath).load_module()
        add_module(package_name, loaded)
        _config_module(loaded, module_settings)
    else:
        imported = importlib.import_module(path)
        _config_module(imported, module_settings)


def load_default_module():
    """
    Load default module.
    """
    for module in ['__init__.py', 'plugin.py']:
        try:
            SourceFileLoader('plugin', module).load_module()
            break
        except FileNotFoundError:
            pass
    else:
        raise LoadException("Unable to find __init__.py or plugin.py.")


def load_modules(modules):
    """
    Load the specified modules.

    If no modules are specified, this function will attempt
    to load __init__.py. If unable to load, it will attempt to load plugin.py.
    If both files cannot be loaded, this function will exit the program.

    If modules are specified, this function will attempt to load all modules
    in the order they are specified. Failure to load a module will exit
    the program.
    """
    if not modules:
        load_default_module()
        return

    for module in modules:
        try:
            load_module(module)
        except FileNotFoundError:
            raise LoadException("File not found: {0}".format(module))



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import unittest
from unittest.mock import patch

from .. import add_parser, parse_all, settings, _parsers
from ..parser import SettingsParser


class TestHelper(unittest.TestCase):
    @patch.dict(_parsers)
    def test_add(self):
        parser = SettingsParser("PARSER_PREFIX")
        add_parser('example', parser, parse_now=False)
        assert _parsers['example'] is parser

    @patch.dict(_parsers)
    @patch.dict(settings)
    @patch.dict(os.environ, {'TEST_VAR': 'yes'})
    def test_add_parse_now(self):
        parser = SettingsParser("PARSER_PREFIX")
        parser.add_setting('test_var')
        add_parser('example', parser)
        assert settings.example.test_var == 'yes'

    @patch.dict(_parsers)
    @patch.dict(settings)
    @patch.dict(os.environ, {'TEST_VAR': 'yes'})
    def test_add_parse_all(self):
        parser = SettingsParser("PARSER_PREFIX")
        add_parser('example', parser, parse_now=False)
        parse_all()
        assert settings.example.test_var == 'yes'



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import json
import logging

import requests
import strict_rfc3339
from werkzeug.routing import Map, Rule
from werkzeug.serving import run_simple
from werkzeug.wrappers import Request, Response

from .. import error
from .common import CommonTransport
from .encoding import _serialize_exc

log = logging.getLogger(__name__)


class PayloadEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            ts = obj.timestamp()
            return strict_rfc3339.timestamp_to_rfc3339_utcoffset(ts)


def send_action(url, payload, timeout=60):
    headers = {'Content-type': 'application/json',
               'Accept': 'application/json'}

    _data = json.dumps(payload, cls=PayloadEncoder)
    return requests.post(url, data=_data, headers=headers, timeout=timeout) \
        .json()


class HttpTransport(CommonTransport):
    """
    HttpTransport implements a transport protocol between skygear and
    plugin that communicates over a HTTP connection.
    """

    def _url_map(self):
        """
        Returns a URL routing map.
        """
        return Map([
            Rule('/', endpoint='_'),
            ])

    def __init__(self, addr, registry=None, debug=False):
        super().__init__(registry)
        self.url_map = self._url_map()
        if ':' in addr:
            hostname, port = addr.split(':', 2)
        else:
            hostname = addr
            port = 8000
        self.hostname = hostname.strip() if hostname.strip() else '0.0.0.0'
        try:
            self.port = int(port)
        except ValueError:
            self.port = 8000
        self.debug = debug

    @Request.application
    def dispatch(self, request):
        """
        Application from the view of the plugin web server.
        """

        # Returns an HTTP response in JSON format
        try:
            output = self._dispatch(request)
        except Exception as e:
            log.exception("exception while handling request")
            output = dict(error=_serialize_exc(e).as_dict())
        return Response(json.dumps(output), mimetype="application/json")

    def _dispatch(self, request):
        """
        Dispatches request to a plugin extension point function.
        """
        kind, name, ctx, param = self.read_request(request)
        if kind == 'init':
            raise Exception('Init trigger is deprecated, '
                            'use init event instead')
        elif kind == 'provider':
            action = param.pop('action')
            return self.call_provider(ctx, name, action, param)
        elif kind == 'handler':
            return self.call_handler(ctx, name, param)
        elif kind == 'event':
            return self.call_event_func(name, param)
        else:
            return self.call_func(ctx, kind, name, param)

    def read_request(self, request):
        """
        Reads and returns HTTP request.
        """
        adapter = self.url_map.bind_to_environ(request.environ)
        _, values = adapter.match()

        request_data = request.get_data(as_text=True)
        req = json.loads(request_data) if request_data else {}

        kind = req.get('kind')
        name = req.get('name')
        param = req.get('param', {})
        ctx = req.get('context', {})

        return kind, name, ctx, param

    def run(self):
        """
        Start the web server.
        """
        run_simple(self.hostname, self.port, self.dispatch,
                   threaded=True,
                   use_reloader=self.debug)

    def send_action(self, action_name, payload, url, timeout):
        resp = send_action(url,
                           payload,
                           timeout=timeout)
        if 'error' in resp:
            raise error.SkygearException.from_dict(resp['error'])
        return resp



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

import configargparse as argparse

# `options` is the py-skygear options read from argv or os.environ.
options = argparse.Namespace()


def add_app_arguments(ap: argparse.ArgumentParser):
    ap.add_argument('--apikey', metavar='APIKEY', action='store',
                    default=None,
                    help="API Key of the application",
                    env_var='API_KEY')
    ap.add_argument('--masterkey', metavar='MASTERKEY', action='store',
                    default=None,
                    help="Master Key of the application",
                    env_var='MASTER_KEY')
    ap.add_argument('--appname', metavar='APPNAME', action='store',
                    default='',
                    help="Application name of the skygear daemon",
                    env_var='APP_NAME')


def add_skygear_arguments(ap: argparse.ArgumentParser):
    ap.add_argument('--skygear-address', metavar='ADDR', action='store',
                    default='tcp://127.0.0.1:5555',
                    help="Binds to this socket for skygear",
                    env_var='SKYGEAR_ADDRESS')
    ap.add_argument('--skygear-endpoint', metavar='ENDPOINT', action='store',
                    default='http://127.0.0.1:3000',
                    help="Send to this addres for skygear handlers",
                    env_var='SKYGEAR_ENDPOINT')
    ap.add_argument('--pubsub-url', action='store', default=None,
                    env_var='PUBSUB_URL',
                    help="The URL of the pubsub server, should start with "
                         "ws:// or wss:// and include the path")


def add_static_asset_arguments(ap: argparse.ArgumentParser):
    ap.add_argument('--collect-assets', metavar='DIST', action='store',
                    help="Collect static assets to a directory")
    ap.add_argument('--force-assets', action='store_true',
                    help="Remove dist folder before proceeding")
    ap.add_argument('--serve-static-assets', action='store_true',
                    env_var='SERVE_STATIC_ASSETS',
                    help="Enable to serve static asset from plugin process")
    ap.add_argument('--ignore-public-html', action='store_true',
                    env_var='IGNORE_PUBLIC_HTML',
                    help="Ignore public_html directory for static assets.")


def add_plugin_arguments(ap: argparse.ArgumentParser):
    ap.add_argument('--subprocess', dest='subprocess', action='store',
                    nargs='+',
                    metavar=('(init|op|hook|handler|timer)', 'name'),
                    help='Trigger subprocess everytime for debug')
    ap.add_argument('--http', action='store_true',
                    help='Trigger http web server',
                    env_var='HTTP')
    ap.add_argument('--http-addr', metavar='HTTP_ADDR', action='store',
                    default='0.0.0.0:8000',
                    help='Address where htp web server listen to',
                    env_var='HTTP_ADDR')
    ap.add_argument('--zmq-thread-pool', metavar='ZMQ_THREAD_POOL',
                    action='store',
                    default=4, type=int,
                    help='Number of thread in ZMQTransport thread pool',
                    env_var='ZMQ_THREAD_POOL')
    ap.add_argument('--zmq-thread-limit', metavar='ZMQ_THREAD_LIMIT',
                    action='store',
                    default=10, type=int,
                    help='Max number of thread in ZMQTransport thread pool',
                    env_var='ZMQ_THREAD_LIMIT')
    ap.add_argument('modules', nargs='*', default=[])  # env_var: LOAD_MODULES


def add_fs_asset_store_arguments(ap: argparse.ArgumentParser):
    ap.add_argument('--asset-store-url-prefix', action='store',
                    metavar='PREFIX', env_var='ASSET_STORE_URL_PREFIX',
                    help='URL prefix of fs asset store (Only applicable for fs \
                          asset store')
    ap.add_argument('--asset-store-secret', action='store', metavar='PATH',
                    env_var='ASSET_STORE_SECRET',
                    help='Secret for signing assets on fs asset store')


def add_s3_asset_store_arguments(ap: argparse.ArgumentParser):
    ap.add_argument('--asset-store-access-key', action='store', metavar='KEY',
                    env_var='ASSET_STORE_ACCESS_KEY',
                    help='Access key for s3 asset store')
    ap.add_argument('--asset-store-secret-key', action='store',
                    metavar='SECRET', env_var='ASSET_STORE_SECRET_KEY',
                    help='Secret key for s3 asset store')
    ap.add_argument('--asset-store-region', action='store', metavar='REGION',
                    env_var='ASSET_STORE_REGION',
                    help='Region for s3 asset store')
    ap.add_argument('--asset-store-bucket', action='store', metavar='BUCKET',
                    env_var='ASSET_STORE_BUCKET',
                    help='Bucket name for s3 asset store')
    ap.add_argument('--asset-store-s3-url-prefix', action='store',
                    metavar='PREFIX', env_var='ASSET_STORE_S3_URL_PREFIX',
                    help='URL prefix for S3 asset store')


def add_cloud_asset_store_arguments(ap: argparse.ArgumentParser):
    ap.add_argument('--cloud-asset-host', action='store', metavar='HOST',
                    env_var='CLOUD_ASSET_HOST',
                    help='Host of cloud asset store')
    ap.add_argument('--cloud-asset-token', action='store', metavar='TOKEN',
                    env_var='CLOUD_ASSET_TOKEN',
                    help='Token of cloud asset store')
    ap.add_argument('--cloud-asset-public-prefix', action='store',
                    metavar='PREFIX', env_var='CLOUD_ASSET_PUBLIC_PREFIX',
                    help='URL prefix of public asset on cloud asset store')
    ap.add_argument('--cloud-asset-private-prefix', action='store',
                    metavar='PREFIX', env_var='CLOUD_ASSET_PRIVATE_PREFIX',
                    help='URL prefix of private asset on cloud asset store')


def add_asset_arguments(ap: argparse.ArgumentParser):
    ap.add_argument('--asset-store', action='store', metavar='(fs|s3|cloud)',
                    default='fs', env_var='ASSET_STORE',
                    help='Type of asset store')
    ap.add_argument('--asset-store-public', action='store_true',
                    help='Make asset public accessible',
                    env_var='ASSET_STORE_PUBLIC')
    add_fs_asset_store_arguments(ap)
    add_s3_asset_store_arguments(ap)
    add_cloud_asset_store_arguments(ap)


def add_logging_arguments(ap: argparse.ArgumentParser):
    ap.add_argument('--loglevel', action='store', default='INFO',
                    help="Log level",
                    env_var='LOG_LEVEL')


def add_debug_arguments(ap: argparse.ArgumentParser):
    ap.add_argument('--debug', action='store_true',
                    help='Enable debugging features',
                    env_var='DEBUG')


def get_argument_parser():
    ap = argparse.ArgumentParser(description='Skygear Plugin runner')
    add_app_arguments(ap)
    add_skygear_arguments(ap)
    add_plugin_arguments(ap)
    add_asset_arguments(ap)
    add_static_asset_arguments(ap)
    add_logging_arguments(ap)
    add_debug_arguments(ap)
    return ap


def _module_name(name):
    if '~' in name and not name.endswith('~py'):
        return ''
    elif name.endswith('~py'):
        return name[:-3]
    else:
        return name


def _parse_load_modules_envvar(val):
    """
    Return a list of modules by parsing the LOAD_MODULES environment
    variable.

    The module name may be suffixed with a runtime specifier such as
    `~js` or `~py`. If a runtime specifier exists and the specified runtime
    is not python, the module will not be loaded.
    """
    if not val:
        return []

    modules = []
    if ':' in val:
        modules = val.split(':')
    elif ',' in val:
        modules = val.split(',')
    else:
        modules = val.split(' ')

    return [_module_name(m) for m in modules if _module_name(m)]


def parse_args():
    global options
    options = get_argument_parser().parse_args(namespace=options)

    # configargparse does not support env_var for positional argument,
    # therefore the LOAD_MODULES env_var is loaded manually.
    if not options.modules:
        options.modules = _parse_load_modules_envvar(os.getenv('LOAD_MODULES'))

    return options



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

from sqlalchemy.schema import Table
from sqlalchemy.sql import text

from skygear.container import SkygearContainer
from skygear.utils import db


class BaseTestCase(unittest.TestCase):
    app_name = 'app+db'

    def setUp(self):
        SkygearContainer.set_default_app_name(self.app_name)
        with db.conn() as conn:
            conn.execute("CREATE SCHEMA IF NOT EXISTS \"app_{0}\""
                         .format(self.app_name))
            conn.execute(
                "set search_path to \"app_{0}\", public;".format(
                    self.app_name
                )
            )
            conn.execute("""
                CREATE TABLE IF NOT EXISTS note (
                    id text PRIMARY KEY,
                    content text
                );""")
            sql = text("""
                INSERT INTO note (id, content)
                VALUES (:id, :content);
                """)
            conn.execute(sql,
                         id='first',
                         content='Hello World!')

    def tearDown(self):
        with db.conn() as conn:
            conn.execute("DROP TABLE \"app_{0}\".note;".format(self.app_name))


class TestResetPassword(BaseTestCase):
    def test_correct_db_context(self):
        with db.conn() as conn:
            result = conn.execute("""
                SELECT id, content FROM note
            """)
            r = result.fetchone()
            assert r['id'] == 'first'
            assert r['content'] == 'Hello World!'


class TestReflectTable(BaseTestCase):
    def test_get_metadata(self):
        metadata = db._get_metadata()
        assert db._get_metadata() == metadata
        assert metadata is not None

    def test_get_table(self):
        table = db.get_table('note')
        assert isinstance(table, Table)
        assert table.c.content is not None

    def test_get_table_nonexistent(self):
        with self.assertRaises(Exception):
            db.get_table('something')

    def test_has_table(self):
        assert db.has_table('note') is True

    def test_has_table_nonexistent(self):
        assert db.has_table('something') is False



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from .common import BaseAssetSigner


class TestBaseAssetSigner(unittest.TestCase):
    def test_init(self):
        assert BaseAssetSigner().public is False
        assert BaseAssetSigner(public=True).public is True

    def test_available(self):
        assert BaseAssetSigner().available() is True
        assert BaseAssetSigner(public=True).available() is True

    def test_signature_required(self):
        assert BaseAssetSigner().signature_required is True
        assert BaseAssetSigner(public=True).signature_required is False



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
import os
import unittest
from unittest.mock import MagicMock, patch

from .. import common
from ...error import SkygearException
from ...models import Record, RecordID
from ...registry import Registry
from ...utils.context import current_context
from ..common import CommonTransport
from ..encoding import deserialize_or_none, serialize_record


class TestCommonTransport(unittest.TestCase):
    def setUp(self):
        self.registry = Registry()
        self.transport = CommonTransport(self.registry)
        self.ctx = {'state': 'happy'}

    def tearDown(self):
        self.registry = None
        self.transport = None

    @patch('skygear.registry.Registry.get_func')
    def testCallFuncGetCorrectObject(self, mocker):
        mocker.return_value = MagicMock()
        self.transport.call_func(self.ctx, 'timer', 'name', {})
        mocker.assert_called_once_with('timer', 'name')
        mocker.return_value.assert_called_once_with()

    @patch('skygear.registry.Registry.get_func')
    def testCallFuncContext(self, mocker):
        def assertState(expectedState):
            def func():
                assert current_context().get('state') == expectedState
            return func

        mocker.return_value = MagicMock(side_effect=assertState("happy"))
        self.transport.call_func(self.ctx, 'timer', 'name', {})
        mocker.return_value.assert_called_once_with()
        assert 'state' not in current_context()

    @patch('skygear.registry.Registry.get_func')
    def testCallFuncResult(self, mocker):
        mocker.return_value = MagicMock(return_value={'data': 'hello'})
        result = self.transport.call_func(self.ctx, 'timer', 'name', {})
        assert result['result'] == {'data': 'hello'}

    @patch('skygear.registry.Registry.get_provider')
    def testCallProviderGetCorrectObject(self, mocker):
        mocker.return_value = MagicMock()
        self.transport.call_provider(self.ctx, 'name', 'action', {})
        mocker.assert_called_once_with('name')
        mocker.return_value.handle_action.assert_called_once_with('action', {})

    @patch('skygear.registry.Registry.get_provider')
    def testCallProviderContext(self, mocker):
        def assertState(expectedState):
            def func():
                assert current_context().get('state') == expectedState
            return func

        mocker.return_value = MagicMock(side_effect=assertState("happy"))
        self.transport.call_provider(self.ctx, 'name', 'action', {})
        mocker.return_value.handle_action.assert_called_once_with('action', {})
        assert 'state' not in current_context()

    @patch('skygear.registry.Registry.get_provider')
    def testCallProviderResult(self, mocker):
        provider_mock = MagicMock()
        provider_mock.handle_action.return_value = {'data': 'hello'}
        mocker.return_value = provider_mock
        result = self.transport.call_provider(self.ctx, 'name', 'action', {})
        assert result['result'] == {'data': 'hello'}

    @patch('skygear.registry.Registry.get_func')
    def testCallFuncGenericException(self, mocker):
        exc = Exception('Error occurred')
        mocker.return_value = MagicMock(side_effect=exc)
        result = self.transport.call_func(self.ctx, 'timer', 'name', {})
        assert result['error']['message'] == 'Error occurred'

    @patch('skygear.registry.Registry.get_func')
    def testCallFuncSkygearException(self, mocker):
        exc = SkygearException('Error occurred', 1, {'data': 'hello'})
        mocker.return_value = MagicMock(side_effect=exc)
        result = self.transport.call_func(self.ctx, 'timer', 'name', {})
        assert result['error']['message'] == 'Error occurred'
        assert result['error']['code'] == 1

    def testOpDictArg(self):
        mock = MagicMock(return_value={'result': 'OK'})
        self.transport.op(mock, dict(named='value'))
        mock.assert_called_with(named='value')

    def testOpArrayArg(self):
        mock = MagicMock(return_value={'result': 'OK'})
        self.transport.op(mock, [1, 2])
        mock.assert_called_with(1, 2)

    def testHandlerWithStrReturn(self):
        mock = MagicMock(return_value='Hello')
        response = self.transport.handler(mock, {
            'path': '/',
            'method': 'POST',
            'header': {'Content-Type': ['text/plain; charset=utf-8']},
            'body': base64.b64encode(b'rawstream')
        })
        assert response['header']['Content-Type'] == \
            ['text/plain; charset=utf-8']
        assert response['status'] == 200
        assert base64.b64decode(response['body']) == b'Hello'

    def testHandlerWithDictReturn(self):
        mock = MagicMock(return_value={
            'hello': 'world'
        })
        response = self.transport.handler(mock, {
            'path': '/',
            'method': 'POST',
            'header': {'Content-Type': ['text/plain; charset=utf-8']},
            'body': base64.b64encode(b'rawstream')
        })
        assert response['header']['Content-Type'] == ['application/json']
        assert response['status'] == 200
        assert base64.b64decode(response['body']) == b'{"hello": "world"}'

    def testHandlerWithResponseReturn(self):
        from werkzeug.utils import redirect

        mock = MagicMock(return_value=redirect('https://skygear.io/'))
        response = self.transport.handler(mock, {
            'path': '/',
            'method': 'POST',
            'header': {'Content-Type': ['text/plain; charset=utf-8']},
            'body': base64.b64encode(b'rawstream')
        })
        assert response['header']['Location'] == ['https://skygear.io/']
        assert response['status'] == 302

    def testHook(self):
        record_id = RecordID('note', 'note1')
        updated_record = Record(record_id, 'owner', None,
                                data={'data': 'updated'})
        old_record = Record(record_id, 'owner', None, data={'data': 'old'})
        new_record = Record(record_id, 'owner', None, data={'data': 'new'})
        param = {
            'original': serialize_record(old_record),
            'record': serialize_record(new_record),
            }
        mock = MagicMock(return_value=updated_record)
        returned_record = self.transport.hook(mock, param)
        assert mock.call_count == 1
        args, kwargs = mock.call_args
        assert args[0].data == new_record.data
        assert args[1].data == old_record.data
        assert deserialize_or_none(returned_record).data == updated_record.data

    def testTimer(self):
        mock = MagicMock()
        self.transport.timer(mock)
        mock.assert_called_once_with()

    def testProvider(self):
        mock = MagicMock()
        self.transport.provider(mock, 'action', {'data': 'hello'})
        mock.handle_action.assert_called_with('action', {'data': 'hello'})


class TestBase64Encoding(unittest.TestCase):
    @patch.dict(os.environ, {'SKYGEAR_CONTEXT': 'e30='})
    def testFromEnviron(self):
        d = common.dict_from_base64_environ('SKYGEAR_CONTEXT')
        assert d == {}

    def testEncoding(self):
        assert common.encode_base64_json({}) == b'e30='

    def testDecoding(self):
        assert common.decode_base64_json('e30=') == {}


class TestHandleException(unittest.TestCase):
    def testHandleException(self):
        exc = Exception()
        assert common.handle_exception(exc) is exc

    def testHandleSkygearException(self):
        exc = SkygearException('Error occurred', 1, {'data': 'hello'})
        assert common.handle_exception(exc) is exc



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import mimetypes
import os
import shutil

from werkzeug.exceptions import NotFound

from . import Response
from .registry import get_registry
from .utils.assets import StaticAssetsLoader

log = logging.getLogger(__name__)
_registry = get_registry()


class CollectorException(Exception):
    pass


class StaticAssetsCollector:
    def __init__(self, dist):
        self.dist = os.path.abspath(dist)

    @property
    def base_path(self):
        return os.path.join(self.dist, 'static')

    def _prefix_path(self, prefix):
        prefix_path = os.path.abspath(os.path.join(self.base_path, prefix))
        if not prefix_path.startswith(self.base_path):
            raise CollectorException(
                'Prefix {} (absolute path {}) does not reside in'
                ' base path {}.'.format(prefix, prefix_path, self.base_path))
        return prefix_path

    def collect(self, prefix, loader):
        if not isinstance(loader, StaticAssetsLoader):
            raise ValueError('The second argument must be an instance '
                             'of StaticAssetsLoader.')
        prefix_path = self._prefix_path(prefix)
        log.debug('Prefix path is %s', prefix_path)
        loader.copy_into(prefix_path)

    def clean(self):
        shutil.rmtree(self.dist)


def serve_static_assets(request, basepath):
    """
    Serve static assets with the given request object.
    """
    if not request.path.startswith(basepath):
        raise ValueError('Request path does not start with basepath ""'
                         .format(basepath))

    path = request.path[len(basepath):]
    try:
        loader, subpath = _registry.get_static_assets(path)
    except KeyError:
        raise NotFound()

    if subpath and subpath[0] == '/':
        subpath = subpath[1:]

    if not loader.exists_asset(subpath):
        log.debug('Asset not found: {}'.format(subpath))
        raise NotFound()

    content_type, _ = mimetypes.guess_type(subpath)
    return Response(loader.get_asset(subpath), content_type=content_type)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from collections import defaultdict

log = logging.getLogger(__name__)


def get_registry():
    return _registry


def _iter_class_hierarchy(klass):
    klasses = [klass]
    while klasses:
        cls = klasses.pop()
        yield cls
        klasses += list(cls.__bases__)


class Registry:
    """Registry holds a mapping of registred functions and their parameters that
    are callable by Skygear plugin system.

    Developers are not expected to create an instance directly. Instead, they
    should use the function get_registry() to the shared instance.
    """

    def __init__(self):
        self.func_map = {
            'op': {},
            'hook': {},
            'timer': {},
            'event': {},
        }
        self.param_map = {
            'op': [],
            'handler': [],
            'hook': [],
            'timer': [],
            'event': [],
            'provider': [],
        }
        self.handler = defaultdict(dict)
        self.providers = {}
        self.static_assets = {}
        self.exception_handlers = {}

    def _add_param(self, kind, param):
        """
        Add a param dict to the registry. If the param already exist
        in the registry, the existing param will be removed. An param
        is a dictionary containing info of the declared extension point.
        """
        param_list = self.param_map[kind]

        # Check existing param. Remove it if the param has the same name
        # as the to-be-added param.
        for i in range(len(param_list)):
            if param_list[i].get('name') == param.get('name'):
                del param_list[i]
                break
        param_list.append(param)

    def register_hook(self, name, func, *args, **kwargs):
        if kwargs['type'] is None:
            raise ValueError("type is required for hook")
        if kwargs['trigger'] is None:
            raise ValueError("trigger is required for hook")
        kwargs['name'] = name

        if name in self.func_map['hook']:
            log.warning("Replacing previously registered hook '%s'.", name)

        self.func_map['hook'][name] = func
        self._add_param('hook', kwargs)

        log.debug("Registered hook '%s' to skygear!", name)

    def register_event(self, name, func, *args, **kwargs):
        if name in self.func_map['event']:
            log.warning("Replacing previously registered event handler '%s'",
                        name)

        self.func_map['event'][name] = func
        self._add_param('event', {
            'name': name
        })

    def register_op(self, name, func, *args, **kwargs):
        if name in self.func_map['op']:
            log.warning("Replacing previously registered lambda '%s'.", name)

        self.func_map['op'][name] = func
        self._add_param('op', {
            'name': name,
            'key_required': kwargs.get('key_required',
                                       kwargs.get('auth_required', False)),
            'user_required': kwargs.get('user_required', False),
        })

        log.debug("Registered lambda '%s' to skygear!", name)

    def register_timer(self, name, func, *args, **kwargs):
        kwargs['name'] = name

        if name in self.func_map['op']:
            log.warning("Replacing previously registered timer '%s'.", name)

        self.func_map['timer'][name] = func
        self._add_param('timer', kwargs)

        log.debug("Registered timer '%s' to skygear!", name)

    def register_handler(self, name, func, *args, **kwargs):
        methods = kwargs.get('method', ['GET', 'POST', 'PUT'])
        if isinstance(methods, str):
            methods = [methods]

        for m in methods:
            if m in self.handler[name]:
                log.warning(
                    "Replacing previously registered handler '%s' (%s).",
                    name,
                    m
                )
            self.handler[name][m] = func
        self._add_param('handler', {
            'name': name,
            'methods': methods,
            'key_required': kwargs.get('key_required', False),
            'user_required': kwargs.get('user_required', False),
        })

        log.debug("Registered handler '%s' (%s) to skygear!",
                  name,
                  ', '.join(methods))

    def register_provider(self, provider_type, provider_id, provider,
                          **kwargs):
        kwargs['type'] = provider_type
        kwargs['id'] = provider_id

        if provider_id in self.providers:
            log.warning("Replacing previously registered provider '%s'.",
                        provider_id)

        self.providers[provider_id] = provider
        self._add_param('provider', kwargs)

        log.debug("Registered provider '%s' to skygear!", provider_id)

    def register_static_assets(self, prefix, func):
        self.static_assets[prefix] = func

    def get_static_assets(self, request_path):
        for prefix, loader in self.static_assets.items():
            if request_path.startswith(prefix):
                subpath = request_path[len(prefix):]
                return loader(), subpath
        raise KeyError('Unable to find static assets loader with '
                       'request_path "{}"'.format(request_path))

    def func_list(self):
        return self.param_map

    def get_func(self, kind, name):
        return self.func_map[kind][name]

    def get_provider(self, name):
        return self.providers[name]

    def get_handler(self, name, method):
        return self.handler[name].get(method, None)

    def get_exception_handler(self, klass):
        for cls in _iter_class_hierarchy(klass):
            if cls in self.exception_handlers:
                return self.exception_handlers[cls]
        return None

    def register_exception_handler(self, klass, func):
        self.exception_handlers[klass] = func


_registry = Registry()



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import json
import os
import unittest
from unittest.mock import ANY, patch

from ...registry import Registry
from ..common import encode_base64_json
from ..console import ConsoleTransport


def environ_dict(context):
    return {'SKYGEAR_CONTEXT': encode_base64_json(context).decode('utf-8')}


class TestConsoleTransport(unittest.TestCase):
    def exec(self, args, data, registry=None):
        out = io.StringIO()
        data = data if isinstance(data, str) else json.dumps(data)
        transport = ConsoleTransport(args,
                                     stdin=io.StringIO(data),
                                     stdout=out,
                                     registry=registry or Registry())
        transport.run()
        out.seek(0)
        return json.loads(out.read())

    @patch.dict(os.environ, environ_dict({'context': 'happy'}))
    @patch('skygear.transmitter.console.ConsoleTransport.call_func')
    def testCallFuncWithData(self, mocker):
        mocker.return_value = {'data': 'hello'}
        output = self.exec(['timer', 'name'], {'data': 'bye'})
        mocker.assert_called_once_with({'context': 'happy'}, 'timer', 'name',
                                       {'data': 'bye'})
        assert output == mocker.return_value

    @patch('skygear.transmitter.console.ConsoleTransport.call_func')
    def testOpInvalidJSON(self, mocker):
        mocker.return_value = {}
        output = self.exec(['op', 'john'], "not-valid")
        assert not mocker.called
        expected_output = {'code': 10000, 'info': {},
                           'message': 'unable to parse JSON string'}
        assert output == expected_output

    @patch('skygear.transmitter.console.ConsoleTransport.call_func')
    def testOpEmpty(self, mocker):
        mocker.return_value = {"data": "hi"}
        output = self.exec(['op', 'john'], "")
        mocker.assert_called_once_with(ANY, 'op', 'john', {})
        assert output == mocker.return_value

    @patch('skygear.transmitter.console.ConsoleTransport.call_func')
    def testOp(self, mocker):
        mocker.return_value = {"data": "hi"}
        output = self.exec(['op', 'john'], {'data': 'bye'})
        mocker.assert_called_once_with(ANY, 'op', 'john', {'data': 'bye'})
        assert output == mocker.return_value

    @patch('skygear.transmitter.console.ConsoleTransport.call_event_func')
    def testEvent(self, mocker):
        mocker.return_value = {"data": "okay-event"}
        output = self.exec(['event', 'okay'], {'data': 'haha'})
        mocker.assert_called_once_with('okay', {'data': 'haha'})
        assert output == mocker.return_value

    @patch('skygear.transmitter.console.ConsoleTransport.init_event_handler')
    def testInitEvent(self, mocker):
        mocker.return_value = {'data': 'hello'}
        output = self.exec(['event', 'init'], {'config': {'hello': 'world'}})
        mocker.assert_called_once_with(config={'hello': 'world'})
        assert output.get('result') == mocker.return_value

    @patch('skygear.transmitter.console.ConsoleTransport.call_func')
    def testHook(self, mocker):
        mocker.return_value = {}
        self.exec(['hook', 'john'], {'data': 'bye'})
        mocker.assert_called_once_with(ANY, 'hook', 'john', {'data': 'bye'})

    @patch('skygear.transmitter.console.ConsoleTransport.call_provider')
    def testProvider(self, mocker):
        mocker.return_value = {}
        self.exec(['provider', 'john', 'work'], {'data': 'bye'})
        mocker.assert_called_once_with(ANY, 'john', 'work', {'data': 'bye'})

    @patch('skygear.transmitter.console.ConsoleTransport.call_func')
    def testTimer(self, mocker):
        mocker.return_value = {}
        self.exec(['timer', 'john'], "")
        mocker.assert_called_once_with(ANY, 'timer', 'john', {})



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
from unittest.mock import call, patch

from .. import importutil


class TestGuessPackageName(unittest.TestCase):
    def test_bare_init(self):
        assert importutil.guess_package_name('__init__.py') == 'plugin'

    def test_outside_relpath(self):
        with self.assertRaises(importutil.LoadException):
            importutil.guess_package_name('../__init__.py')

    def test_current_directory(self):
        assert importutil.guess_package_name('.') == 'plugin'

    @patch('os.path.isfile')
    def test_file_path_init(self, mocker):
        mocker.return_value = True
        assert importutil.guess_package_name('./abc/__init__.py') \
            == 'abc'

    @patch('os.path.isfile')
    def test_file_path_other(self, mocker):
        mocker.return_value = True
        assert importutil.guess_package_name('./abc/def.py') \
            == 'abc.def'

    @patch('os.path.isfile')
    @patch('os.path.isdir')
    def test_dir_path(self, dirmocker, filemocker):
        filemocker.return_value = False
        dirmocker.return_value = True
        assert importutil.guess_package_name('./abc/def') \
            == 'abc.def'

    def test_non_existent_file(self):
        with self.assertRaises(importutil.LoadException):
            importutil.guess_package_name('./abc/non_existent.py')


class TestLoadModule(unittest.TestCase):
    def setUp(self):
        self.guess = patch('skygear.importutil.guess_package_name',
                           lambda p: 'guess_package_name')
        self.guess.start()

    def tearDown(self):
        self.guess.stop()

    @patch('skygear.importutil.SourceFileLoader')
    def test_load_module_file(self, mocker):
        with patch('os.path.isfile', lambda p: True):
            importutil.load_module('abc/__init__.py')
        mocker.assert_called_once_with('guess_package_name',
                                       'abc/__init__.py')

    @patch('skygear.importutil.SourceFileLoader')
    def test_load_module_dir(self, mocker):
        with patch('os.path.isdir', lambda p: True):
            importutil.load_module('abc/def')
        mocker.assert_called_once_with('guess_package_name',
                                       'abc/def/__init__.py')

    @patch('importlib.import_module')
    def test_load_module_package(self, mocker):
        importutil.load_module('abc.def')
        mocker.assert_called_once_with('abc.def')


class TestLoadDefaultModule(unittest.TestCase):
    @patch('skygear.importutil.SourceFileLoader')
    def test_load_success(self, mocker):
        importutil.load_default_module()
        mocker.assert_called_once_with('plugin', '__init__.py')

    @patch('skygear.importutil.SourceFileLoader')
    def test_load_not_found(self, mocker):
        mocker.side_effect = FileNotFoundError
        with self.assertRaises(importutil.LoadException):
            importutil.load_default_module()
        mocker.assert_has_calls([
            call('plugin', '__init__.py'),
            call('plugin', 'plugin.py'),
            ])


class TestLoadModules(unittest.TestCase):
    @patch('skygear.importutil.load_default_module')
    def test_empty(self, mocker):
        importutil.load_modules([])
        mocker.assert_called_once_with()

    @patch('skygear.importutil.load_module')
    def test_multiple(self, mocker):
        importutil.load_modules(['abc', 'def'])
        mocker.assert_has_calls([
            call('abc'),
            call('def'),
            ])



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest.mock import MagicMock, patch

from configargparse import Namespace

from .s3 import S3AssetSigner


class TestS3AssetSigner(unittest.TestCase):
    @property
    def mock_options(self):
        return Namespace(asset_store_access_key='mock_s3_access_key',
                         asset_store_secret_key='mock_s3_secret_key',
                         asset_store_region='mock-s3-region',
                         asset_store_bucket='mock-s3-bucket',
                         asset_store_s3_url_prefix=None,
                         asset_store_public=False)

    @patch('skygear.asset.s3.aws_client')
    def test_create(self, mock_aws_client):
        signer = S3AssetSigner.create(self.mock_options)
        assert signer.bucket == 'mock-s3-bucket'
        assert signer.region == 'mock-s3-region'
        assert signer.signature_required is True
        assert signer.url_prefix is None
        mock_aws_client.assert_called_once_with(
            's3',
            aws_access_key_id='mock_s3_access_key',
            aws_secret_access_key='mock_s3_secret_key',
            region_name='mock-s3-region')

    def test_create_fail(self):
        with self.assertRaises(Exception):
            S3AssetSigner.create(
                Namespace(asset_store_access_key=None,
                          asset_store_secret_key='mock_s3_secret_key',
                          asset_store_region='mock-s3-region',
                          asset_store_bucket='mock-s3-bucket',
                          asset_store_public=False))
        with self.assertRaises(Exception):
            S3AssetSigner.create(
                Namespace(asset_store_access_key='mock_s3_access_key',
                          asset_store_secret_key=None,
                          asset_store_region='mock-s3-region',
                          asset_store_bucket='mock-s3-bucket',
                          asset_store_public=False))
        with self.assertRaises(Exception):
            S3AssetSigner.create(
                Namespace(asset_store_access_key='mock_s3_access_key',
                          asset_store_secret_key='mock_s3_secret_key',
                          asset_store_region=None,
                          asset_store_bucket='mock-s3-bucket',
                          asset_store_public=False))
        with self.assertRaises(Exception):
            S3AssetSigner.create(
                Namespace(asset_store_access_key='mock_s3_access_key',
                          asset_store_secret_key='mock_s3_secret_key',
                          asset_store_region='mock-s3-region',
                          asset_store_bucket=None,
                          asset_store_public=False))

    @patch('skygear.asset.s3.aws_client')
    def test_init(self, mock_aws_client):
        signer = S3AssetSigner(access_key='mock_s3_access_key',
                               access_secret='mock_s3_secret_key',
                               region='mock-s3-region',
                               bucket='mock-s3-bucket')
        assert signer.bucket == 'mock-s3-bucket'
        assert signer.region == 'mock-s3-region'
        assert signer.signature_required is True
        mock_aws_client.assert_called_once_with(
            's3',
            aws_access_key_id='mock_s3_access_key',
            aws_secret_access_key='mock_s3_secret_key',
            region_name='mock-s3-region')

    @patch('skygear.asset.s3.aws_client')
    def test_signing(self, mock_aws_client):
        mock_aws_client_instance = MagicMock()
        mock_aws_client.return_value = mock_aws_client_instance
        mock_aws_client_instance.generate_presigned_url.return_value = \
            'http://skygear.dev/signed_url'

        signer = S3AssetSigner.create(self.mock_options)
        assert signer.sign('index.html') == \
            mock_aws_client_instance.generate_presigned_url.return_value
        mock_aws_client_instance.generate_presigned_url.\
            assert_called_once_with('get_object',
                                    Params={
                                        'Bucket': 'mock-s3-bucket',
                                        'Key': 'index.html'
                                    },
                                    ExpiresIn=900)

    def test_signing_public(self):
        options = self.mock_options
        options.asset_store_public = True
        signer = S3AssetSigner.create(options)
        assert signer.sign('index.html') == (
            'https://s3-mock-s3-region.amazonaws.com/'
            'mock-s3-bucket/index.html')

    def test_signing_public_with_url_prefix(self):
        options = self.mock_options
        options.asset_store_public = True
        options.asset_store_s3_url_prefix = 'http://skygear.dev'
        signer = S3AssetSigner.create(options)
        assert signer.sign('index.html') == 'http://skygear.dev/index.html'



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest
from datetime import datetime, timedelta
from unittest.mock import MagicMock, Mock, patch

from configargparse import Namespace
from strict_rfc3339 import timestamp_to_rfc3339_utcoffset

from .cloud import CloudAssetSigner, CloudAssetSignerToken


class TestCloudAssetSignerToken(unittest.TestCase):
    def test_create(self):
        info = {'value': 'mock-token-value',
                'expired_at': '2016-12-08T08:38:33Z',
                'extra': 'mock-token-extra'}
        token = CloudAssetSignerToken.create(info)
        assert token.value == 'mock-token-value'
        assert token.extra == 'mock-token-extra'
        assert token.expired_at == datetime.fromtimestamp(1481186313)

    def test_create_fail(self):
        with self.assertRaises(Exception):
            CloudAssetSignerToken.create({
                'expired_at': '2016-12-08T08:38:33Z',
                'extra': 'mock-token-extra'})

        with self.assertRaises(Exception):
            CloudAssetSignerToken.create({
                'value': 'mock-token-value',
                'extra': 'mock-token-extra'})

        with self.assertRaises(Exception):
            CloudAssetSignerToken.create({
                'value': 'mock-token-value',
                'expired_at': 'invalid-format',
                'extra': 'mock-token-extra'})

    def test_init(self):
        token = CloudAssetSignerToken('mock-token-value',
                                      datetime.fromtimestamp(1481186313),
                                      'mock-token-extra')
        assert token.value == 'mock-token-value'
        assert token.extra == 'mock-token-extra'
        assert token.expired_at == datetime.fromtimestamp(1481186313)

    def test_expiration(self):
        token1 = CloudAssetSignerToken('mock-token-value',
                                       datetime.now() + timedelta(minutes=10),
                                       'mock-token-extra')
        assert token1.expired() is False

        token2 = CloudAssetSignerToken('mock-token-value',
                                       datetime.now() - timedelta(minutes=10),
                                       'mock-token-extra')
        assert token2.expired() is True


class TestCloudAssetSigner(unittest.TestCase):
    def mock_options(self, public=False):
        return Namespace(
            appname='skygear-test',
            cloud_asset_host='http://mock-cloud-asset.dev',
            cloud_asset_token='mock-cloud-asset-token',
            cloud_asset_public_prefix='http://mock-cloud-asset.dev/public',
            cloud_asset_private_prefix='http://mock-cloud-asset.dev/private',
            asset_store_public=public)

    def setUp(self):
        self.request_patcher = patch('skygear.asset.cloud.request')
        self.mock_request = self.request_patcher.start()
        self.mock_request.return_value = MagicMock()
        self.mock_request.return_value.text = json.dumps({
            'value': 'mock-token-value',
            'expired_at': '2016-12-08T08:38:33Z',
            'extra': 'mock-token-extra'})

    def tearDown(self):
        self.request_patcher.stop()
        self.mock_request = None

    def test_create(self):
        signer1 = CloudAssetSigner.create(self.mock_options())
        assert signer1.app_name == 'skygear-test'
        assert signer1.host == 'http://mock-cloud-asset.dev'
        assert signer1.token == 'mock-cloud-asset-token'
        assert signer1.url_prefix == 'http://mock-cloud-asset.dev/private'
        assert signer1.signer_token.value == 'mock-token-value'
        assert signer1.signer_token.extra == 'mock-token-extra'
        assert signer1.signer_token.expired_at == \
            datetime.fromtimestamp(1481186313)

        signer2 = CloudAssetSigner.create(self.mock_options(public=True))
        assert signer2.app_name == 'skygear-test'
        assert signer2.host == 'http://mock-cloud-asset.dev'
        assert signer2.token == 'mock-cloud-asset-token'
        assert signer2.url_prefix == 'http://mock-cloud-asset.dev/public'
        assert signer2.signer_token.value == 'mock-token-value'
        assert signer2.signer_token.extra == 'mock-token-extra'
        assert signer2.signer_token.expired_at == \
            datetime.fromtimestamp(1481186313)

    def test_create_fail(self):
        with self.assertRaises(Exception):
            CloudAssetSigner.create(Namespace(
                appname=None,
                cloud_asset_host='http://mock-cloud-asset.dev',
                cloud_asset_token='mock-cloud-asset-token',
                cloud_asset_public_prefix='http://mock-cloud-asset.dev/public',
                cloud_asset_private_prefix=(
                    'http://mock-cloud-asset.dev/private'),
                asset_store_public=False))

        with self.assertRaises(Exception):
            CloudAssetSigner.create(Namespace(
                appname='skygear-test',
                cloud_asset_host=None,
                cloud_asset_token='mock-cloud-asset-token',
                cloud_asset_public_prefix='http://mock-cloud-asset.dev/public',
                cloud_asset_private_prefix=(
                    'http://mock-cloud-asset.dev/private'),
                asset_store_public=False))

        with self.assertRaises(Exception):
            CloudAssetSigner.create(Namespace(
                appname='skygear-test',
                cloud_asset_host='http://mock-cloud-asset.dev',
                cloud_asset_token=None,
                cloud_asset_public_prefix='http://mock-cloud-asset.dev/public',
                cloud_asset_private_prefix=(
                    'http://mock-cloud-asset.dev/private'),
                asset_store_public=False))

        with self.assertRaises(Exception):
            CloudAssetSigner.create(Namespace(
                appname='skygear-test',
                cloud_asset_host='http://mock-cloud-asset.dev',
                cloud_asset_token='mock-cloud-asset-token',
                cloud_asset_public_prefix='http://mock-cloud-asset.dev/public',
                cloud_asset_private_prefix=None,
                asset_store_public=False))

        with self.assertRaises(Exception):
            CloudAssetSigner.create(Namespace(
                appname='skygear-test',
                cloud_asset_host='http://mock-cloud-asset.dev',
                cloud_asset_token='mock-cloud-asset-token',
                cloud_asset_public_prefix=None,
                cloud_asset_private_prefix=(
                    'http://mock-cloud-asset.dev/private'),
                asset_store_public=True))

    def test_init(self):
        signer = CloudAssetSigner('skygear-test',
                                  'http://mock-cloud-asset.dev',
                                  'mock-cloud-asset-token',
                                  'http://mock-cloud-asset.dev/private')
        assert signer.app_name == 'skygear-test'
        assert signer.host == 'http://mock-cloud-asset.dev'
        assert signer.token == 'mock-cloud-asset-token'
        assert signer.url_prefix == 'http://mock-cloud-asset.dev/private'
        assert signer.signer_token.value == 'mock-token-value'
        assert signer.signer_token.extra == 'mock-token-extra'
        assert signer.signer_token.expired_at == \
            datetime.fromtimestamp(1481186313)

    def test_signer_availability(self):
        self.mock_request.return_value.text = json.dumps({
            'value': 'mock-token-value',
            'expired_at': timestamp_to_rfc3339_utcoffset(
                (datetime.now() + timedelta(minutes=10)).timestamp()),
            'extra': 'mock-token-extra'
            })
        signer1 = CloudAssetSigner.create(self.mock_options())
        assert signer1.available() is True

        self.mock_request.return_value.text = json.dumps({
            'value': 'mock-token-value',
            'expired_at': timestamp_to_rfc3339_utcoffset(
                (datetime.now() - timedelta(minutes=10)).timestamp()),
            'extra': 'mock-token-extra'
            })
        signer2 = CloudAssetSigner.create(self.mock_options())
        assert signer2.available() is False

    @patch('skygear.asset.cloud.datetime')
    def test_refresh_signer_token(self, mock_datetime):
        fixed_now = datetime.now()
        mock_datetime.now.return_value = fixed_now
        CloudAssetSigner.create(self.mock_options())
        self.mock_request.assert_called_once_with(
            'GET', 'http://mock-cloud-asset.dev/token/skygear-test',
            headers={'Authorization': 'Bearer mock-cloud-asset-token'},
            params={
                'expired_at':
                    str(int((fixed_now + timedelta(hours=2)).timestamp()))
            })

    def test_refresh_signer_token_fail(self):
        self.mock_request.return_value.text = json.dumps({
                'Error': 'Testing Error'
            })
        with self.assertRaises(Exception):
            CloudAssetSigner.create(self.mock_options())

    @patch('skygear.asset.cloud.datetime')
    @patch('skygear.asset.cloud.CloudAssetSignerToken.expired',
           Mock(return_value=True))
    def test_sign(self, mock_datetime):
        mock_datetime.now.return_value = datetime.fromtimestamp(1481095934)
        signer = CloudAssetSigner.create(self.mock_options())
        assert signer.sign('index.html') == (
            'http://mock-cloud-asset.dev/private/skygear-test/index.html'
            '?expired_at=1481096834'
            '&signature=peQtnmSFdoQWtFAk3cwLkM3lUspBkIhl5SPlR5hjFm4%3D'
            '.mock-token-extra')

    def test_sign_public(self):
        signer = CloudAssetSigner.create(Namespace(
            appname='skygear-test',
            cloud_asset_host='http://mock-cloud-asset.dev',
            cloud_asset_token='mock-cloud-asset-token',
            cloud_asset_public_prefix='http://mock-cloud-asset.dev/public',
            cloud_asset_private_prefix='http://mock-cloud-asset.dev/private',
            asset_store_public=True))
        assert signer.sign('index.html') == \
            'http://mock-cloud-asset.dev/public/skygear-test/index.html'



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os.path
import signal
import sys

from . import commands
from .container import SkygearContainer
from .importutil import LoadException, load_modules
from .options import parse_args
from .settings import parse_all as parse_all_settings
from .transmitter import ConsoleTransport, HttpTransport, ZmqTransport

log = logging.getLogger(__name__)


def main():
    options = parse_args()
    setup_logging(options)
    if options.collect_assets:
        load(options)
        parse_all_settings()
        commands.collect_static_assets()
    else:
        run_plugin(options)


def load(options):
    # If the directory `public_html` exists in the current directory,
    # assume the user want to publish its content as static assets.
    auto_assets_dir = os.path.abspath('public_html')
    if os.path.exists(auto_assets_dir) and not options.ignore_public_html:
        from .decorators import static_assets

        @static_assets('')
        def auto_assets():
            return auto_assets_dir

    # Create handler for serving static assets.
    if options.serve_static_assets:
        STATIC_ASSETS_PREFIX = 'static'
        from .decorators import handler
        from .assets import serve_static_assets

        @handler('{}/'.format(STATIC_ASSETS_PREFIX))
        @handler(STATIC_ASSETS_PREFIX)
        def work(request):
            return serve_static_assets(request,
                                       '/{}/'.format(STATIC_ASSETS_PREFIX))

    try:
        load_modules(options.modules)
    except LoadException as ex:
        log.error(str(ex))
        sys.exit(1)


def run_plugin(options):
    SkygearContainer.set_default_app_name(options.appname)
    SkygearContainer.set_default_endpoint(options.skygear_endpoint)
    SkygearContainer.set_default_apikey(options.apikey)

    load(options)
    parse_all_settings()
    log.debug("Install signal handler for SIGTERM")
    signal.signal(signal.SIGTERM, sigterm_handler)

    if options.subprocess is not None:
        transport = ConsoleTransport(options.subprocess)
    elif options.http:
        transport = HttpTransport(options.http_addr, debug=options.debug)
    else:
        log.info(
            "Connecting to address %s" % options.skygear_address)
        transport = ZmqTransport(options.skygear_address,
                                 threading=options.zmq_thread_pool,
                                 limit=options.zmq_thread_limit)
    SkygearContainer.set_default_transport(transport)
    transport.run()


def setup_logging(options):
    # TODO: Make it load a stadard python logging config.
    logger = logging.getLogger()
    level = {
        'DEBUG': logging.DEBUG,
        'INFO': logging.INFO,
        'WARN': logging.WARN,
        'ERROR': logging.ERROR,
        'CRITICAL': logging.CRITICAL
    }.get(options.loglevel, logging.INFO)
    logger.setLevel(level)
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('''\
%(asctime)s %(levelname)-5.5s \
[%(name)s:%(lineno)s][%(threadName)s] %(message)s\
''')
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def sigterm_handler(signum, frame):
    """
    When docker stops the container, it sends PID 1 with SIGTERM.
    This function exits the program when SIGTERM is received.

    Remember not to run this program in a shell as PID 1 because the shell may
    not forward the signal.
    """
    sys.exit(1)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from collections import namedtuple

from . import Namespace

SettingItem = namedtuple('SettingItem',
                         'name default atype env_var resolve required')


class SettingsParser:
    """
    SettingsParser allows cloudcode and plugins to define the settings to
    parse. It reads from os.environ and parse the values into an object.

    By default, if a setting called `SOME_VAR` is defined, it will
    first check for `PREFIX_SOME_VAR` and then `SOME_VAR`, stopping at
    the first available variable.
    """
    def __init__(self, prefix):
        """
        Create a new SettingsParser.

        Specify a prefix for environment variable.
        """
        self.prefix = prefix
        self.settings = {}

    def add_setting(self,
                    name,
                    default=None,
                    atype=str,
                    env_var=None,
                    resolve=True,
                    required=True):
        """
        Add a setting to the parser.

        * name - The name of the setting, the value of the setting
          will be stored in the object with attribute of the same name.
        * default - If no environment variable is defined during paring,
          the default value will be used instead. Default is `None`.
        * atype - The type of the value. Specify a type or a function
          that accepts a single argument of the input and return the converted
          value. Default is `str`.
        * env_var - Alternatively name of the environment variable. If not
          specified, the default is the uppercase of the setting name.
        * resolve - In addition to the prefixed environment variable,
          also check the unprefixed variant. Default is `true`.
        * required - Whether the setting is required. The parser will
          raise an exception during parsing if the corresponding environment
          variables are not set. (Empty value is considered "set").
        """
        if not name:
            raise Exception("Setting name must be specified.")

        if name in self.settings:
            raise Exception("Setting named '{}' already defined.".format(name))

        if not env_var:
            env_var = name.upper()

        setting = SettingItem(name, default, atype, env_var, resolve, required)
        self.settings[name] = setting

    def parse_settings(self, namespace=None):
        """
        Parse the settings from environment variables. The settings
        are returned as an object.

        * namespace - Optionally specify a namespace where the settings
          are saved. This object is returned if specified. A new object
          is created if none is specified.
        """
        if not namespace:
            namespace = Namespace()

        for name, setting in self.settings.items():
            setattr(namespace, name, self._parse_setting(setting))
        return namespace

    def _parse_setting(self, setting):
        order = self._resolve_order(setting)
        for var_name in order:
            if var_name in os.environ:
                val = os.environ[var_name]
                break
        else:
            val = setting.default

        if val is None and setting.required:
            raise Exception("Setting named \"{}\" is defined "
                            "but it is not set.".format(setting.name))

        if val is not None:
            val = setting.atype(val)

        return val

    def _resolve_order(self, setting):
        order = ["{}_{}".format(self.prefix, setting.env_var)]
        if setting.resolve:
            order.append(setting.env_var)
        return order



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import importlib
from argparse import Namespace

from .parser import SettingsParser
from .module import _config_module, get_module


_parsers = {}
"""
`settings` is namespaced settings for plugins. Following is an example of how
a plugin developer can register settings and query back the value.


Declaring the required settings:
```
from skygear.settings import SettingsParser, add_parser

parser = SettingsParser('SKYGEAR_CMS')
parser.add_setting('prefix', default='cms')
parser.add_setting('static_assets_prefix', default='/static/:prefix/')

add_parser('cms', parser)
```

The settings module will read the environment and set the value. For example,
os.environ `SKYGEAR_CMS_PREFIX` will be available as follows:

```
from skygear.settings import settings

settings.cms.prefix
```
"""
settings = Namespace()


def add_parser(name, parser, parse_now=True):
    """
    Add the specified SettingsParser. If parse_now is True (the default),
    the parser is run immediately.
    """
    global _parsers
    if name in _parsers:
        raise Exception("Parser \"{}\" already defined.", name)
    _parsers[name] = parser
    if parse_now:
        _parse(name, parser)


def _parse(name, parser):
    global settings

    ns = getattr(settings, name, Namespace())
    ns = parser.parse_settings(ns)
    setattr(settings, name, ns)


def parse_all():
    """
    Parse all settings.
    """
    global _parsers
    global settings

    for name, parser in _parsers.items():
        _parse(name, parser)
    return settings


def config_module(name, *args, **kwargs):
    """
    Try to config the already imported module on boot time. If the package is
    not already imported in boot time, will try to import as normal package
    and config it.

    config_module will not import another copy of the module if it was already
    loaded at boot time.

    To config a module, the `includeme` function will be called and all
    skygear lambda functions, database hooks, etc. are expected to be declared
    in the `includeme` function as follows:

    ```
    import skygear

    def includeme(settings, *args, **kwargs):
        @skygear.op('some:lambda')
        def some_lambda_func():
            return {
                'success': True,
                'message': 'Some message being returned'
            }

        @skygear.after_save('some_record', async=True)
        def some_record_after_save(record, original_record, db):
            return {
                'success': True
            }

        # other lambda functions

    ```

    The `includeme` function will be called in the following cases:

    1. The module is declared on Skygear Plugin Runner, i.e. execute
       `py-skygear some_module`.

    2. The module is configured in cloud code, i.e. user calls
       `skygear.config('some_module')` in his cloud code.

    When the `includeme` function is called, settings will be passed as an
    argument.

    """
    global settings
    try:
        module = get_module(name)
    except NameError:
        module = importlib.import_module(name)
    _config_module(module, settings, *args, **kwargs)


__all__ = [
    Namespace,
    SettingsParser, settings,
    config_module,
    add_parser, parse_all
]



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
from unittest.mock import MagicMock

from .. import action


class TestPushNotification(unittest.TestCase):
    def setUp(self):
        self.mock_container = MagicMock()

    def tearDown(self):
        self.mock_container = None

    def test_push_device(self):
        action.push_device(self.mock_container,
                           'device01',
                           {'apns': {'alert': 'hello'}})
        self.mock_container.send_action\
            .assert_called_once_with('push:device', {
                'device_ids': ['device01'],
                'notification': {'apns': {'alert': 'hello'}}
            })

    def test_push_devices(self):
        action.push_devices(self.mock_container,
                            ['device01', 'device02', 'device03'],
                            {'apns': {'alert': 'hello'}})
        self.mock_container.send_action\
            .assert_called_once_with('push:device', {
                'device_ids': ['device01', 'device02', 'device03'],
                'notification': {'apns': {'alert': 'hello'}}
            })

    def test_push_device_with_topic(self):
        action.push_device(self.mock_container,
                           'device01',
                           {'apns': {'alert': 'hello'}},
                           topic='io.skygear.example.app')
        self.mock_container.send_action\
            .assert_called_once_with('push:device', {
                'topic': 'io.skygear.example.app',
                'device_ids': ['device01'],
                'notification': {'apns': {'alert': 'hello'}}
            })

    def test_push_devices_with_topic(self):
        action.push_devices(self.mock_container,
                            ['device01', 'device02', 'device03'],
                            {'apns': {'alert': 'hello'}},
                            topic='io.skygear.example.app')
        self.mock_container.send_action\
            .assert_called_once_with('push:device', {
                'topic': 'io.skygear.example.app',
                'device_ids': ['device01', 'device02', 'device03'],
                'notification': {'apns': {'alert': 'hello'}}
            })

    def test_push_user(self):
        action.push_user(self.mock_container,
                         'user01',
                         {'apns': {'alert': 'hello'}})
        self.mock_container.send_action\
            .assert_called_once_with('push:user', {
                'user_ids': ['user01'],
                'notification': {'apns': {'alert': 'hello'}}
            })

    def test_push_users(self):
        action.push_users(self.mock_container,
                          ['user01', 'user02', 'user03'],
                          {'apns': {'alert': 'hello'}})
        self.mock_container.send_action\
            .assert_called_once_with('push:user', {
                'user_ids': ['user01', 'user02', 'user03'],
                'notification': {'apns': {'alert': 'hello'}}
            })

    def test_push_user_with_topic(self):
        action.push_user(self.mock_container,
                         'user01',
                         {'apns': {'alert': 'hello'}},
                         topic='io.skygear.example.app')
        self.mock_container.send_action\
            .assert_called_once_with('push:user', {
                'topic': 'io.skygear.example.app',
                'user_ids': ['user01'],
                'notification': {'apns': {'alert': 'hello'}}
            })

    def test_push_users_with_topic(self):
        action.push_users(self.mock_container,
                          ['user01', 'user02', 'user03'],
                          {'apns': {'alert': 'hello'}},
                          topic='io.skygear.example.app')
        self.mock_container.send_action\
            .assert_called_once_with('push:user', {
                'topic': 'io.skygear.example.app',
                'user_ids': ['user01', 'user02', 'user03'],
                'notification': {'apns': {'alert': 'hello'}}
            })



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


loaded_modules = {}


def _config_module(module, settings, *args, **kwargs):
    try:
        module.includeme(settings, *args, **kwargs)
    except AttributeError as e:
        msg = "Module {} does not implement function includeme"\
            .format(module.__name__)
        raise NotImplementedError(msg)


def add_module(name, module):
    loaded_modules[name] = module


def get_module(name):
    try:
        return loaded_modules[name]
    except KeyError as ex:
        raise NameError("name '{}' is not defined".format(name))



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configargparse as argparse
from boto3 import client as aws_client

from .common import BaseAssetSigner


class S3AssetSigner(BaseAssetSigner):
    def __init__(self, access_key: str, access_secret: str,
                 region: str, bucket: str,
                 url_prefix: str = None, public: bool = False):
        super().__init__(public)
        self.bucket = bucket
        self.region = region
        self.url_prefix = url_prefix
        self.client = aws_client('s3',
                                 aws_access_key_id=access_key,
                                 aws_secret_access_key=access_secret,
                                 region_name=region)

    def public_url(self, name: str) -> str:
        if self.url_prefix:
            return '/'.join([self.url_prefix, name])
        return 'https://s3-{}.amazonaws.com/{}/{}'.format(self.region,
                                                          self.bucket,
                                                          name)

    def sign(self, name: str) -> str:
        if not self.signature_required:
            return self.public_url(name)

        params = {'Bucket': self.bucket, 'Key': name}
        expire_duration = int(self.signature_expiry_duration.total_seconds())
        return self.client.generate_presigned_url('get_object',
                                                  Params=params,
                                                  ExpiresIn=expire_duration)

    @classmethod
    def create(cls, options: argparse.Namespace) -> BaseAssetSigner:
        access_key = options.asset_store_access_key
        if not access_key:
            raise Exception('Missing access key for s3 asset store')

        access_secret = options.asset_store_secret_key
        if not access_secret:
            raise Exception('Missing access secret for s3 asset store')

        region = options.asset_store_region
        if not region:
            raise Exception('Missing region for s3 asset store')

        bucket = options.asset_store_bucket
        if not bucket:
            raise Exception('Missing bucket name for s3 asset store')

        return cls(access_key, access_secret, region, bucket,
                   options.asset_store_s3_url_prefix,
                   options.asset_store_public)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import hashlib
import hmac
from datetime import datetime

import configargparse as argparse

from .common import BaseAssetSigner


class FileSystemAssetSigner(BaseAssetSigner):
    def __init__(self, url_prefix: str, secret: str, public: bool = False):
        super().__init__(public)
        self.url_prefix = url_prefix
        self.secret = secret

    def sign(self, name: str) -> str:
        if not self.signature_required:
            return '{}/{}'.format(self.url_prefix, name)

        expired_at = datetime.now() + self.signature_expiry_duration
        expired_at_str = str(int(expired_at.timestamp()))

        hasher = hmac.new(self.secret.encode('utf-8'),
                          digestmod=hashlib.sha256)
        hasher.update(name.encode('utf-8'))
        hasher.update(expired_at_str.encode('utf-8'))

        signature = base64\
            .urlsafe_b64encode(hasher.digest()[:-1])\
            .decode('utf-8')

        return '{}/{}?expiredAt={}&signature={}'.format(self.url_prefix,
                                                        name,
                                                        expired_at_str,
                                                        signature)

    @classmethod
    def create(cls, options: argparse.Namespace) -> BaseAssetSigner:
        url_prefix = options.asset_store_url_prefix
        if not url_prefix:
            raise Exception('Missing URL prefix of fs asset store')

        secret = options.asset_store_secret
        if not secret:
            raise Exception('Missing signing secret for fs asset store')

        return cls(url_prefix, secret, options.asset_store_public)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os.path
import shutil
import tempfile
import unittest
from unittest.mock import patch

from werkzeug.test import EnvironBuilder
from werkzeug.wrappers import Request

from .. import assets
from ..utils.assets import DictStaticAssetsLoader, DirectoryStaticAssetsLoader


class TestStaticAssetsCollector(unittest.TestCase):
    def setUp(self):
        self.dist = tempfile.mkdtemp()
        self.collector = assets.StaticAssetsCollector(self.dist)

    def tearDown(self):
        if os.path.exists(self.dist):
            shutil.rmtree(self.dist)

    def test_clean(self):
        self.collector.clean()
        assert os.path.exists(self.dist) is False

    def test_collect(self):
        tmp_dir = tempfile.mkdtemp()
        try:
            with open(os.path.join(tmp_dir, 'index.txt'), 'w') as f:
                f.write('Hello World!')

            self.collector.collect('hello-world',
                                   DirectoryStaticAssetsLoader(tmp_dir))
        finally:
            shutil.rmtree(tmp_dir)

        collected_path = os.path.join(self.collector.base_path,
                                      'hello-world',
                                      'index.txt')
        with open(collected_path, 'r') as f:
            assert f.read() == 'Hello World!'

    def test_collect_incorrect_prefix(self):
        tmp_dir = tempfile.mkdtemp()
        try:
            with open(os.path.join(tmp_dir, 'index.txt'), 'w') as f:
                f.write('Hello World!')

            with self.assertRaises(assets.CollectorException):
                self.collector.collect('../hello-world',
                                       DirectoryStaticAssetsLoader(tmp_dir))
        finally:
            shutil.rmtree(tmp_dir)


class TestServeStaticAssets(unittest.TestCase):
    def setUp(self):
        assets = {
                'content/index.html': 'Hello World!'
                }
        self.loader = DictStaticAssetsLoader(assets)

    @patch('skygear.registry.Registry.get_static_assets')
    def test_response(self, mocker):
        mocker.return_value = (self.loader, 'content/index.html')
        builder = EnvironBuilder(method='GET',
                                 path='/static/myloader/content/index.html')
        request = Request(builder.get_environ())
        response = assets.serve_static_assets(request, '/static/')
        assert response.data == b'Hello World!'
        mocker.assert_called_once_with('myloader/content/index.html')



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
NotAuthenticated = 101
PermissionDenied = 102
AccessKeyNotAccepted = 103
AccessTokenNotAccepted = 104
InvalidCredentials = 105
InvalidSignature = 106
BadRequest = 107
InvalidArgument = 108
Duplicated = 109
ResourceNotFound = 110
NotSupported = 111
NotImplemented = 112
ConstraintViolated = 113
IncompatibleSchema = 114
AtomicOperationFailure = 115
PartialOperationFailure = 116
UndefinedOperation = 117
PluginUnavailable = 118
PluginTimeout = 119
RecordQueryInvalid = 120
PluginInitializing = 121

UnexpectedError = 10000


class SkygearException(Exception):
    """
    An exception object to be raised by plugin to indicate that
    the requested operation is not successful, or that the operation
    is aborted for whatever reason.
    """
    def __init__(self, message, code=UnexpectedError, info=None):
        self.message = message
        self.code = code
        self.info = info or {}
        super(SkygearException, self).__init__(message)

    def as_dict(self):
        return {
            'message': self.message,
            'code': self.code,
            'info': self.info,
            }

    def readable_message(self):
        """
        This method returns a user readable string for the error.
        """
        # Make exception of the complexity check here because we need to
        # cover each error case.
        # pylama:ignore=C901
        # #lizard forgives the complexity
        if self.code == NotAuthenticated:
            return "You have to be authenticated to perform this operation."
        elif (self.code in [PermissionDenied, AccessKeyNotAccepted,
                            AccessTokenNotAccepted]):
            return "You are not allowed to perform this operation."
        elif self.code == InvalidCredentials:
            return "You are not allowed to log in because the credentials " + \
                "you provided are not valid."
        elif self.code in [InvalidSignature, BadRequest]:
            return "The server is unable to process the request."
        elif self.code == InvalidArgument:
            return "The server is unable to process the data."
        elif self.code == Duplicated:
            return "This request contains duplicate of an existing " + \
                "resource on the server."
        elif self.code == ResourceNotFound:
            return "The requested resource is not found."
        elif self.code == NotSupported:
            return "This operation is not supported."
        elif self.code == NotImplemented:
            return "This operation is not implemented."
        elif (self.code in [ConstraintViolated, IncompatibleSchema,
                            AtomicOperationFailure, PartialOperationFailure]):
            return "A problem occurred while processing this request."
        elif self.code == UndefinedOperation:
            return "The requested operation is not available."
        elif self.code in [PluginInitializing, PluginUnavailable]:
            return "The server is not ready yet."
        elif self.code == PluginTimeout:
            return "The server took too long to process."
        elif self.code == RecordQueryInvalid:
            return "A problem occurred while processing this request."
        return "An unexpected error has occurred."

    @classmethod
    def from_dict(cls, error_dict):
        message = error_dict.get('message', "An error has occurred.")
        code = error_dict.get('code', UnexpectedError)
        info = error_dict.get('info', {})
        return cls(message, code, info)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .decorators import (
    hook,
    handler,
    op,
    event,
    every,
    provides,
    before_save,
    after_save,
    before_delete,
    after_delete,
    rest,
    static_assets,
    exception_handler,
)

from .restful import RestfulRecord, RestfulResource
from .utils.db import conn as db_conn
from .utils.http import Response
from .settings import config_module as config

__all__ = [hook, handler, op, event, every, provides,
           before_save, after_save, before_delete, after_delete,
           rest, RestfulRecord, RestfulResource,
           static_assets, exception_handler,
           config,
           db_conn,
           Response]



import sys
from os import path

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

README = path.abspath(path.join(path.dirname(__file__), 'README.md'))

classifiers = [
    'License :: OSI Approved :: Apache Software License',
    'Intended Audience :: Developers',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Operating System :: POSIX',
    'Operating System :: MacOS :: MacOS X',
    'Environment :: Web Environment',
    'Development Status :: 3 - Alpha',
]

class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['skygear']
        self.test_suite = True
    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)

extras_require={
    'zmq': ['pyzmq>=14.7'],
}

setup(
      name='skygear',
      version='1.0.0',
      packages=find_packages(),
      description='Python plugin runtime for Skygear',
      long_description=open(README).read(),
      classifiers=classifiers,
      author='Rick Mak',
      author_email='rick.mak@gmail.com',
      url='https://github.com/SkygearIO/py-skygear',
      license='Apache License, Version 2.0',
      install_requires=[
            'psycopg2>=2.6.1',
            'SQLAlchemy>=1.0.8',
            'strict-rfc3339>=0.5',
            'requests',
            'websocket-client>=0.32.0',
            'bcrypt==2.0.0',
            'ConfigArgParse>=0.12.0',
            'werkzeug>=0.11.0',
            'boto3>=1.4',
      ],
      extras_require=extras_require,
      cmdclass= {'test': PyTest},
      tests_require=[
            'pytest',
      ] + extras_require['zmq'],
      entry_points={
          'console_scripts': [
              'py-skygear = skygear.bin:main'
          ]
      },
)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import hashlib
import hmac
from datetime import datetime, timedelta
from json.decoder import JSONDecoder
from urllib.parse import quote as percent_encode

import configargparse as argparse
from requests import request
from strict_rfc3339 import InvalidRFC3339Error, rfc3339_to_timestamp

from .common import BaseAssetSigner


class CloudAssetSignerToken:
    def __init__(self, value: str, expired_at: datetime, extra: str = ''):
        self._value = value
        self._expired_at = expired_at
        self._extra = extra

    @property
    def value(self) -> str:
        return self._value

    @property
    def expired_at(self) -> datetime:
        return self._expired_at

    @property
    def extra(self) -> str:
        return self._extra

    def expired(self) -> bool:
        expiry_duration = datetime.now() - self.expired_at
        return expiry_duration.total_seconds() > 0

    @classmethod
    def create(cls, info: dict) -> object:
        value = info.get('value')
        if not value:
            raise Exception('Missing signer token value for cloud asset')

        expired_at_str = info.get('expired_at')
        if not expired_at_str:
            raise Exception('Missing expiry time for cloud asset signer token')

        expire_timestamp = None
        try:
            expire_timestamp = rfc3339_to_timestamp(expired_at_str)
        except InvalidRFC3339Error as ex:
            raise Exception(
                'Invalid format of expiry time for cloud asset signer token')

        return cls(value,
                   datetime.fromtimestamp(expire_timestamp),
                   info.get('extra'))


class CloudAssetSigner(BaseAssetSigner):
    def __init__(self, app_name: str, host: str, token: str,
                 url_prefix: str, public: bool = False):
        super().__init__(public)
        self.app_name = app_name
        self.host = host
        self.token = token
        self.url_prefix = url_prefix
        self.signer_token = None
        self.refresh_signer_token()

    @property
    def signer_token_expiry_duration(self):
        return timedelta(hours=2)

    def available(self) -> bool:
        return self.signer_token and not self.signer_token.expired()

    def refresh_signer_token(self):
        url = "{}/token/{}".format(self.host, self.app_name)
        authorization_token = 'Bearer {}'.format(self.token)
        expired_at = datetime.now() + self.signer_token_expiry_duration
        expired_at_str = str(int(expired_at.timestamp()))
        resp = request('GET', url,
                       headers={'Authorization': authorization_token},
                       params={'expired_at': expired_at_str})

        resp_dict = None
        try:
            resp_dict = JSONDecoder().decode(resp.text)
        except Exception as ex:
            raise Exception(
                'Fail to decode the response from cloud asset')

        resp_err = resp_dict.get('Error')
        if resp_err:
            raise Exception('Fail to get the signer token')

        self.signer_token = CloudAssetSignerToken.create(resp_dict)

    def sign(self, name: str) -> str:
        if not self.available():
            self.refresh_signer_token()

        url = '/'.join([self.url_prefix, self.app_name, name])
        if not self.signature_required:
            return url

        expired_at = datetime.now() + self.signature_expiry_duration
        expired_at_str = str(int(expired_at.timestamp()))

        hasher = hmac.new(self.signer_token.value.encode('utf-8'),
                          digestmod=hashlib.sha256)
        for each_info in [self.app_name, name, expired_at_str,
                          self.signer_token.extra]:
            hasher.update(each_info.encode('utf-8'))

        # use standard_b64encode instead of urlsafe_b64encode following
        # the implementation of cloud asset
        signature = base64\
            .standard_b64encode(hasher.digest())\
            .decode('utf-8')
        signature_and_extra = percent_encode(
            '{}.{}'.format(signature, self.signer_token.extra))

        return '{}?expired_at={}&signature={}'.format(url, expired_at_str,
                                                      signature_and_extra)

    @classmethod
    def create(cls, options: argparse.Namespace) -> BaseAssetSigner:
        app_name = options.appname
        if not app_name:
            raise Exception('Missing app name')

        host = options.cloud_asset_host
        if not host:
            raise Exception('Missing host for cloud asset store')

        token = options.cloud_asset_token
        if not token:
            raise Exception('Missing token for cloud asset store')

        public = options.asset_store_public
        url_prefix = options.cloud_asset_public_prefix if public else \
            options.cloud_asset_private_prefix

        if not url_prefix:
            raise Exception('Missing url prefix for cloud asset')

        return cls(app_name, host, token, url_prefix, public)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
from unittest.mock import ANY, call, patch

from werkzeug.wrappers import Request

from .. import decorators as d
from ..registry import Registry  # noqa
from ..utils.assets import DirectoryStaticAssetsLoader


class TestOpDecorators(unittest.TestCase):
    @patch('skygear.registry.Registry.register_op')
    def test_register(self, mocker):
        @d.op('do:something')
        def fn():
            pass
        mocker.assert_called_with(
            'do:something', ANY)


class TestTimerDecorators(unittest.TestCase):
    @patch('skygear.registry.Registry.register_timer')
    def test_register(self, mocker):
        @d.every('* * * * * *')
        def fn():
            pass
        mocker.assert_called_with(
            'skygear.tests.test_decorators.fn', ANY,
            spec='* * * * * *')


class TestHookDecorators(unittest.TestCase):
    @patch('skygear.registry.Registry.register_hook')
    def test_before_save(self, mocker):
        @d.before_save('note')
        def fn():
            pass
        mocker.assert_called_with(
            'skygear.tests.test_decorators.fn', ANY,
            type='note', trigger='beforeSave')

    @patch('skygear.registry.Registry.register_hook')
    def test_after_save(self, mocker):
        @d.after_save('note')
        def fn():
            pass
        mocker.assert_called_with(
            'skygear.tests.test_decorators.fn', ANY,
            type='note', trigger='afterSave')

    @patch('skygear.registry.Registry.register_hook')
    def test_before_delete(self, mocker):
        @d.before_delete('note')
        def fn():
            pass
        mocker.assert_called_with(
            'skygear.tests.test_decorators.fn', ANY,
            type='note', trigger='beforeDelete')

    @patch('skygear.registry.Registry.register_hook')
    def test_after_delete(self, mocker):
        @d.after_delete('note')
        def fn():
            pass
        mocker.assert_called_with(
            'skygear.tests.test_decorators.fn', ANY,
            type='note', trigger='afterDelete')

    @patch('skygear.registry.Registry.register_hook')
    def test_hook(self, mocker):
        @d.hook('beforeSave', type="note")
        def fn():
            pass
        mocker.assert_called_with(
            'skygear.tests.test_decorators.fn', ANY,
            type='note', trigger='beforeSave')


def test_fix_handler_path():
    assert d._fix_handler_path('hello:world') == 'hello/world'
    assert d._fix_handler_path('hello/world') == 'hello/world'
    assert d._fix_handler_path('/hello/world/') == 'hello/world'


class TestRestDecorator(unittest.TestCase):
    @patch('skygear.registry.Registry.register_handler')
    def test_register(self, mock):
        @d.rest('/hello/world', user_required=True)
        class HelloWorld:
            def handle_request(self, base_name, request):
                pass

        mock.assert_has_calls([
            call('hello/world/', ANY, method=ANY,
                 user_required=True),
            call('hello/world', ANY, method=ANY,
                 user_required=True),
            ])

        registered_func = mock.call_args[0][1]
        with patch.object(HelloWorld, 'handle_request') as handle_request:
            req = Request({})
            registered_func(req)
            handle_request.assert_called_with('/hello/world', req)


class TestStaticAssetsDecorator(unittest.TestCase):
    @patch('skygear.registry.Registry.register_static_assets')
    def test_register(self, mock):
        @d.static_assets('admin')
        def fn():
            return '/tmp/public'

        mock.assert_called_with('admin', ANY)
        loader = fn()
        assert isinstance(loader, DirectoryStaticAssetsLoader)
        assert loader.dirpath == '/tmp/public'

    @patch('skygear.registry.Registry.register_static_assets')
    def test_register_with_slash(self, mock):
        @d.static_assets('/admin')
        def fn():
            return '/tmp/public'

        mock.assert_called_with('admin', ANY)


class TestExceptionHandlerDecorator(unittest.TestCase):
    @patch('skygear.registry.Registry.register_exception_handler')
    def test_register(self, mock):
        class SomeException(Exception):
            pass

        @d.exception_handler(SomeException)
        def fn(exc):
            return exc

        mock.assert_called_with(SomeException, ANY)
        exc = fn(SomeException())
        assert isinstance(exc, SomeException)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import unittest
from unittest.mock import patch

from .. import Namespace, SettingsParser


class TestSettingsParser(unittest.TestCase):
    def setUp(self):
        self.parser = SettingsParser('PARSER_PREFIX')

    def tearDown(self):
        self.parser = None

    @patch.dict(os.environ, {'TEST_VAR': 'yes'})
    def test_var(self):
        self.parser.add_setting('test_var')
        assert self.parser.parse_settings().test_var == 'yes'

    def test_var_with_default(self):
        self.parser.add_setting('test_var', default='yes')
        assert self.parser.parse_settings().test_var == 'yes'

    def test_var_unspecified(self):
        self.parser.add_setting('test_var')
        with self.assertRaises(Exception):
            self.parser.parse_settings().test_var

    def test_var_unspecified_ignored(self):
        self.parser.add_setting('test_var', required=False)
        assert self.parser.parse_settings().test_var is None

    @patch.dict(os.environ, {'ANOTHER_VAR': 'yes'})
    def test_var_explicit(self):
        self.parser.add_setting('test_var', env_var='ANOTHER_VAR')
        assert self.parser.parse_settings().test_var == 'yes'

    @patch.dict(os.environ, {'PARSER_PREFIX_TEST_VAR': 'no',
                             'TEST_VAR': 'yes'})
    def test_var_correct_resolve(self):
        self.parser.add_setting('test_var')
        assert self.parser.parse_settings().test_var == 'no'

    @patch.dict(os.environ, {'TEST_VAR': 'yes'})
    def test_var_unresolved(self):
        self.parser.add_setting('test_var', resolve=False)
        with self.assertRaises(Exception):
            self.parser.parse_settings().test_var

    @patch.dict(os.environ, {'TEST_VAR': '42'})
    def test_var_int(self):
        self.parser.add_setting('test_var', atype=int)
        assert self.parser.parse_settings().test_var == 42

    def test_var_existing_namespace(self):
        self.parser.add_setting('test_var', default='yes')
        ns = Namespace()
        setattr(ns, 'existing', 'no')
        assert self.parser.parse_settings(ns) is ns
        assert ns.existing == 'no'
        assert ns.test_var == 'yes'



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import unittest
from unittest.mock import ANY, patch

from werkzeug.test import Client
from werkzeug.wrappers import BaseResponse

from ...registry import Registry
from ..common import encode_base64_json
from ..http import HttpTransport


def headers_with_context(data):
    return {'X-Skygear-Plugin-Context': encode_base64_json(data)}


class TestHttpTransport(unittest.TestCase):
    def get_app(self):
        return self.transport.dispatch

    def get_client(self, dispatcher=None):
        return Client(dispatcher or self.get_app(), BaseResponse)

    def setUp(self):
        self.transport = HttpTransport('127.0.0.1:8888', Registry())

    def tearDown(self):
        self.transport = None

    @patch('skygear.transmitter.http.run_simple')
    def testRun(self, mocker):
        self.transport.run()
        mocker.assert_called_once_with('127.0.0.1', 8888,
                                       self.transport.dispatch,
                                       threaded=True,
                                       use_reloader=False)

    @patch('skygear.transmitter.http.HttpTransport.call_func')
    def testCallFuncWithData(self, mocker):
        mocker.return_value = {'data': 'hello'}
        data = {
            'context': {'context': 'happy'},
            'kind': 'timer',
            'name': 'name',
            'param': {
                'data': 'bye'
            }
        }
        resp = self.get_client().post('/', data=json.dumps(data))
        assert resp.status_code == 200
        mocker.assert_called_once_with({'context': 'happy'}, 'timer', 'name',
                                       {'data': 'bye'})
        assert json.loads(resp.get_data(as_text=True)) == mocker.return_value

    @patch('skygear.transmitter.http.HttpTransport.call_func')
    def testOp(self, mocker):
        mocker.return_value = {}
        data = {
            'kind': 'op',
            'name': 'john'
        }
        resp = self.get_client().post('/', data=json.dumps(data))
        assert resp.status_code == 200
        mocker.assert_called_once_with(ANY, 'op', 'john', ANY)

    @patch('skygear.transmitter.http.HttpTransport.call_event_func')
    def testEvent(self, mocker):
        mocker.return_value = {}
        data = {
            'kind': 'event',
            'name': 'funny'
        }
        resp = self.get_client().post('/', data=json.dumps(data))
        assert resp.status_code == 200
        mocker.assert_called_once_with('funny', ANY)

    @patch('skygear.transmitter.http.HttpTransport.init_event_handler')
    def testInitEvent(self, mocker):
        mocker.return_value = {'data': 'hello'}
        data = {
            'kind': 'event',
            'name': 'init',
            'param': {
                'config': {'hello': 'world'}
            }
        }
        transport = HttpTransport('127.0.0.1:8888', Registry())
        client = self.get_client(transport.dispatch)
        resp = client.post('/', data=json.dumps(data))

        assert resp.status_code == 200
        mocker.assert_called_once_with(config={'hello': 'world'})

        resp_data = json.loads(resp.get_data(as_text=True))
        assert resp_data.get('result') == mocker.return_value

    @patch('skygear.transmitter.http.HttpTransport.call_func')
    def testHook(self, mocker):
        mocker.return_value = {}
        data = {
            'kind': 'hook',
            'name': 'john'
        }
        resp = self.get_client().post('/', data=json.dumps(data))
        assert resp.status_code == 200
        mocker.assert_called_once_with(ANY, 'hook', 'john', ANY)

    @patch('skygear.transmitter.http.HttpTransport.call_provider')
    def testProvider(self, mocker):
        mocker.return_value = {}
        data = {
            'kind': 'provider',
            'name': 'john',
            'param': {
                'action': 'work'
            }
        }
        resp = self.get_client().post('/', data=json.dumps(data))
        assert resp.status_code == 200
        mocker.assert_called_once_with(ANY, 'john', 'work', ANY)
        args, kwargs = mocker.call_args

    @patch('skygear.transmitter.http.HttpTransport.call_func')
    def testTimer(self, mocker):
        mocker.return_value = {}
        data = {
            'kind': 'timer',
            'name': 'john'
        }
        resp = self.get_client().post('/', data=json.dumps(data))
        assert resp.status_code == 200
        mocker.assert_called_once_with(ANY, 'timer', 'john', ANY)

    @patch('skygear.transmitter.http.HttpTransport.call_handler')
    def testHandler(self, mocker):
        mocker.return_value = {}
        data = {
            'kind': 'handler',
            'name': 'apple/pie'
        }
        resp = self.get_client().post('/', data=json.dumps(data))
        assert resp.status_code == 200
        mocker.assert_called_once_with(ANY, 'apple/pie', ANY)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from datetime import datetime
from unittest.mock import patch

from configargparse import Namespace

from .fs import FileSystemAssetSigner


class TestFileSystemAssetSigner(unittest.TestCase):
    @property
    def mock_options(self):
        return Namespace(asset_store_url_prefix='http://skygear.dev/files',
                         asset_store_secret='asset_secret',
                         asset_store_public=False)

    def test_create(self):
        signer = FileSystemAssetSigner.create(self.mock_options)
        assert signer.url_prefix == 'http://skygear.dev/files'
        assert signer.secret == 'asset_secret'
        assert signer.signature_required is True

    def test_create_fail(self):
        with self.assertRaises(Exception):
            FileSystemAssetSigner.create(
                Namespace(asset_store_url_prefix=None,
                          asset_store_secret='asset_secret',
                          asset_store_public=False))
        with self.assertRaises(Exception):
            FileSystemAssetSigner.create(
                Namespace(asset_store_url_prefix='http://skygear.dev/files',
                          asset_store_secret=None,
                          asset_store_public=False))

    def test_init(self):
        signer = FileSystemAssetSigner(url_prefix='http://skygear.dev/files',
                                       secret='asset_secret')
        assert signer.url_prefix == 'http://skygear.dev/files'
        assert signer.secret == 'asset_secret'
        assert signer.signature_required is True

    @patch('skygear.asset.fs.datetime')
    def test_signing(self, mock_datetime):
        mock_datetime.now.return_value = datetime.fromtimestamp(1481095934)
        signer = FileSystemAssetSigner.create(self.mock_options)
        assert signer.sign('index.html') == (
            'http://skygear.dev/files/index.html'
            '?expiredAt=1481096834'
            '&signature=R5kMq2neUkCGBjQD6zSv99PajRvI0EqMesuRHQS4hA==')

    def test_signing_public(self):
        options = self.mock_options
        options.asset_store_public = True
        signer = FileSystemAssetSigner.create(options)
        assert signer.sign('index.html') \
            == 'http://skygear.dev/files/index.html'



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

from ..options import _parse_load_modules_envvar


class TestParseLoadModules(unittest.TestCase):
    def test_colon(self):
        self.assertEqual(_parse_load_modules_envvar('a:b'), ['a', 'b'])

    def test_comma(self):
        self.assertEqual(_parse_load_modules_envvar('a,b'), ['a', 'b'])

    def test_space(self):
        self.assertEqual(_parse_load_modules_envvar('a b'), ['a', 'b'])

    def test_empty(self):
        self.assertEqual(_parse_load_modules_envvar(''), [])

    def test_ignore(self):
        self.assertEqual(_parse_load_modules_envvar('a~js,b,c~py'), ['b', 'c'])



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import threading
import time
import unittest

import zmq

from ...transmitter.zmq import (HEARTBEAT_INTERVAL, HEARTBEAT_LIVENESS,
                                PPP_HEARTBEAT, PPP_REQUEST, PPP_RESPONSE,
                                ZmqTransport)


class TestZmq(unittest.TestCase):
    def test_numebr_of_thread_startup(self):
        context = zmq.Context()
        transport = ZmqTransport('tcp://0.0.0.0:12345',
                                 context=context,
                                 threading=5)
        transport.start()
        self.assertEqual(len(transport.threads), 5)
        for t in transport.threads:
            self.assertEqual(t.is_alive(), True)
        transport.stop()
        for t in transport.threads:
            self.assertEqual(t.is_alive(), False)
        context.destroy()

    def test_worker_dead(self):
        context = zmq.Context()
        t = threading.Thread(target=dead_router,
                             args=(context, 'tcp://0.0.0.0:23456', 3,))
        t.start()
        transport = ZmqTransport('tcp://0.0.0.0:23456',
                                 context=context,
                                 threading=3)
        transport.start()
        transport.stop()
        time.sleep(HEARTBEAT_INTERVAL * HEARTBEAT_LIVENESS * 2)
        alive_threads = [_t for _t in transport.threads if _t.is_alive()]
        self.assertEqual(len(alive_threads), 0)
        self.assertEqual(transport.threads_opened, 3)
        t.join()
        context.destroy()

    def test_maintain_worker_count(self):
        context = zmq.Context()
        t = threading.Thread(target=dead_router,
                             args=(context, 'tcp://0.0.0.0:34567', 3,))
        t.start()
        transport = ZmqTransport('tcp://0.0.0.0:34567',
                                 context=context,
                                 threading=3)
        transport.start()
        transport_t = threading.Thread(
            target=transport.maintain_workers_count)
        transport_t.start()
        time.sleep(HEARTBEAT_INTERVAL * HEARTBEAT_LIVENESS * 2)
        alive_threads = [_t for _t in transport.threads if _t.is_alive()]
        self.assertEqual(len(alive_threads), 3)
        transport.stop()
        transport_t.join()
        self.assertEqual(transport_t.is_alive(), False)
        alive_threads = [_t for _t in transport.threads if _t.is_alive()]
        self.assertEqual(len(alive_threads), 0)
        self.assertEqual(transport.threads_opened, 6)
        t.join()
        context.destroy()

    @unittest.mock.patch('skygear.transmitter.zmq.Worker.handle_message')
    def test_spawn_new_worker(self, handle_message_mock):
        def slow_response(*args, **kwargs):
            time.sleep(HEARTBEAT_INTERVAL * HEARTBEAT_LIVENESS * 2)
            return b'{}'
        handle_message_mock.side_effect = slow_response

        context = zmq.Context()
        t = threading.Thread(target=request_router,
                             args=(context,
                                   'tcp://0.0.0.0:23456',
                                   {'name': 'foo'}))
        t.start()
        transport = ZmqTransport('tcp://0.0.0.0:23456',
                                 context=context,
                                 threading=1)
        transport.start()
        transport_t = threading.Thread(
            target=transport.maintain_workers_count)
        transport_t.start()
        time.sleep(HEARTBEAT_INTERVAL * HEARTBEAT_LIVENESS * 2)
        self.assertEqual(len(transport.threads), 2)
        self.assertEqual(transport.threads_opened, 2)
        handle_message_mock.assert_called_with(b'{"name": "foo"}', {
            'bounce_count': 0,
            'request_id': 'REQ-ID',
        })
        t.join()
        transport.stop()
        transport_t.join()
        context.destroy()

    def test_one_off_worker(self):
        context = zmq.Context()
        expected_body = {'method': 'POST', 'payload': {'key': 'value'}}
        return_value = '{"result": {"response_key": "response_value"}}'
        t = threading.Thread(target=response_router,
                             args=(context,
                                   'tcp://0.0.0.0:23456',
                                   expected_body,
                                   return_value))
        t.start()
        transport = ZmqTransport('tcp://0.0.0.0:23456',
                                 context=context,
                                 threading=1)
        transport.start()
        result = transport.send_action('action_name', {'key': 'value'})
        self.assertEqual(
            result,
            {'result': {'response_key': 'response_value'}}
        )
        t.join()
        context.destroy()


def dead_router(context, addr, count):
    """
    This router will send malformed frame that crash the worker
    """
    router = context.socket(zmq.ROUTER)
    router.bind(addr)
    i = 0
    while i < count:
        router.poll()
        frames = router.recv_multipart()
        if not frames:
            break
        frames.extend([b'good', b'bye'])
        router.send_multipart(frames)
        i = i + 1
    router.close()


def response_router(context, addr, expected_body, response):
    """
    This router will send predefined response body to the worker
    """
    router = context.socket(zmq.ROUTER)
    router.bind(addr)
    while True:
        router.poll()
        frames = router.recv_multipart()
        if len(frames) == 2:
            router.send(PPP_HEARTBEAT)
            continue
        body = frames[7].decode('utf8')
        assert expected_body == json.loads(body)

        frames[3] = PPP_RESPONSE
        frames[7] = response.encode('utf8')
        router.send_multipart(frames)
        router.close()
        break


def request_router(context, addr, body):
    """
    This router will send predefined request body to the worker
    """
    router = context.socket(zmq.ROUTER)
    router.bind(addr)
    router.poll()
    frames = router.recv_multipart()
    router.send(PPP_HEARTBEAT)

    address = frames[0]
    frames = [
        address,
        address,
        b'',
        PPP_REQUEST,
        b'0',
        b'REQ-ID',
        b'',
        json.dumps(body).encode('utf8')
    ]
    router.send_multipart(frames)

    router.poll()
    router.recv_multipart()
    router.close()
